from typing import Union
from pathlib import Path

import importlib
import json
from pathlib import Path
from xml.etree import ElementTree as ET
from collections import OrderedDict
from qtpy import QtGui
from qtpy.QtCore import QDateTime, QTime
from pymodaq_gui.parameter import Parameter

from pyqtgraph.parametertree.Parameter import PARAM_TYPES, PARAM_NAMES


def walk_parameters_to_xml(parent_elt=None, param=None):
    """
        To convert a parameter object (and children) to xml data tree.

        =============== ================================ ==================================
        **Parameters**   **Type**                         **Description**

        *parent_elt*     XML element                      the root element
        *param*          instance of pyqtgraph parameter  Parameter object to be converted
        =============== ================================ ==================================

        Returns
        -------
        XML element : parent_elt
            XML element with subelements from Parameter object

        See Also
        --------
        add_text_to_elt, walk_parameters_to_xml, dict_from_param

    """

    if type(param) is None:
        raise TypeError('No valid param input')

    if parent_elt is None:
        opts = dict_from_param(param)
        parent_elt = ET.Element(param.name(), **opts)
        param_type = str(param.type())
        if 'group' not in param_type:  # covers 'group', custom 'groupmove', 'groupai' ...
            add_text_to_elt(parent_elt, param)

    params_list = param.children()
    for param in params_list:

        opts = dict_from_param(param)
        elt = ET.Element(param.name(), **opts)
        param_type = str(param.type())
        if 'group' not in param_type:  # covers 'group', custom 'groupmove'...
            add_text_to_elt(elt, param)
        else:
            walk_parameters_to_xml(elt, param)

        parent_elt.append(elt)
    return parent_elt


def add_text_to_elt(elt, param):
    """Add a text filed in a xml element corresponding to the parameter value

    Parameters
    ----------
    elt: XML elt
    param: Parameter

    See Also
    --------
    add_text_to_elt, walk_parameters_to_xml, dict_from_param
    """
    param_type = str(param.type())
    if 'bool' in param_type or 'led' in param_type:
        if param.value():
            text = '1'
        else:
            text = '0'
    elif param_type == 'itemselect':
        if param.value() is not None:
            elt.set('all_items',
                    str(param.value()['all_items']))  # use list(eval(val_str[1:-1])) to get back a list of strings
            text = str(param.value()['selected'])  # use list(eval(val_str[1:-1])) to get back a list of strings
        else:
            text = str(None)
    elif param_type == 'color':
        text = str([param.value().red(), param.value().green(), param.value().blue(), param.value().alpha()])
    elif param_type == 'list':
        if isinstance(param.value(), str):
            text = "str('{}')".format(param.value())
        elif isinstance(param.value(), int):
            text = 'int({})'.format(param.value())
        elif isinstance(param.value(), float):
            text = 'float({})'.format(param.value())
        else:
            text = str(param.value())
    elif param_type == 'int':
        if param.value() is True:  # known bug is True should be clearly specified here
            val = 1
        else:
            val = param.value()
        text = str(val)
    elif param_type == 'date_time':
        text = str(param.value().toMSecsSinceEpoch())
    elif param_type == 'date':
        text = str(QDateTime(param.value(), QTime()).toMSecsSinceEpoch())
    elif param_type == 'table_view':
        try:
            data = dict(classname=param.value().__class__.__name__,
                        module=param.value().__class__.__module__,
                        data=param.value().get_data_all(),
                        header=param.value().header)
            text = json.dumps(data)
        except Exception:
            text = ''
    else:
        text = str(param.value())
    elt.text = text


def dict_from_param(param):
    """Get Parameter properties as a dictionary

    Parameters
    ----------
    param: Parameter

    Returns
    -------
    opts: dict

    See Also
    --------
    add_text_to_elt, walk_parameters_to_xml, dict_from_param
    """
    opts = dict([])
    param_type = str(param.type())
    opts.update(dict(type=param_type))
    title = param.opts['title']
    if title is None:
        title = param.name()
    opts.update(dict(title=title))

    visible = '1'
    if 'visible' in param.opts:
        if param.opts['visible']:
            visible = '1'
        else:
            visible = '0'

    opts.update(dict(visible=visible))

    removable = '1'
    if 'removable' in param.opts:
        if param.opts['removable']:
            removable = '1'
        else:
            removable = '0'
    opts.update(dict(removable=removable))

    readonly = '0'
    if 'readonly' in param.opts:
        if param.opts['readonly']:
            readonly = '1'
        else:
            readonly = '0'
    opts.update(dict(readonly=readonly))

    # if 'limits' in param.opts:
    #     values = str(param.opts['limits'])
    #     opts.update(dict(values=values))

    if 'limits' in param.opts:
        limits = str(param.opts['limits'])
        opts.update(dict(limits=limits))
        # opts.update(dict(values=limits))

    if 'addList' in param.opts:
        addList = str(param.opts['addList'])
        opts.update(dict(addList=addList))

    if 'addText' in param.opts:
        addText = str(param.opts['addText'])
        opts.update(dict(addText=addText))

    if 'detlist' in param.opts:
        detlist = str(param.opts['detlist'])
        opts.update(dict(detlist=detlist))

    if 'movelist' in param.opts:
        movelist = str(param.opts['movelist'])
        opts.update(dict(movelist=movelist))

    if 'show_pb' in param.opts:
        if param.opts['show_pb']:
            show_pb = '1'
        else:
            show_pb = '0'
        opts.update(dict(show_pb=show_pb))

    if 'filetype' in param.opts:
        if param.opts['filetype']:
            filetype = '1'
        else:
            filetype = '0'
        opts.update(dict(filetype=filetype))

    return opts


def elt_to_dict(el):
    """Convert xml element attributes to a dictionnary

    Parameters
    ----------
    el

    Returns
    -------

    """
    param = dict([])

    # name=el.tag, title=title, type=param_type, value=param_value, values=[param_value],
    #              visible=visible, removable=removable, readonly=readonly, show_pb=show_pb)
    param.update(dict(name=el.tag))
    param_type = el.get('type')
    param.update(dict(type=param_type))

    title = el.get('title')
    if title == 'None':
        title = el.tag
    param.update(dict(title=title))

    if 'visible' not in el.attrib.keys():
        visible = True
    else:
        visible = bool(int(el.get('visible')))
    param.update(dict(visible=visible))

    if 'removable' not in el.attrib.keys():
        removable = False
    else:
        removable = bool(int(el.get('removable')))
    param.update(dict(removable=removable))

    if 'readonly' not in el.attrib.keys():
        readonly = False
    else:
        readonly = bool(int(el.get('readonly')))
    param.update(dict(readonly=readonly))

    if 'show_pb' in el.attrib.keys():
        show_pb = bool(int(el.get('show_pb')))
    else:
        show_pb = False
    param.update(dict(show_pb=show_pb))

    if 'filetype' in el.attrib.keys():
        filetype = bool(int(el.get('filetype')))
        param.update(dict(filetype=filetype))

    if 'detlist' in el.attrib.keys():
        detlist = eval(el.get('detlist'))
        param.update(dict(detlist=detlist))

    if 'movelist' in el.attrib.keys():
        movelist = eval(el.get('movelist'))
        param.update(dict(movelist=movelist))

    if 'addList' in el.attrib.keys():
        addList = eval(el.get('addList'))
        param.update(dict(addList=addList))

    if 'addText' in el.attrib.keys():
        addText = str(el.get('addText'))
        param.update(dict(addText=addText))

    # if 'limits' in el.attrib.keys():
    #     try:
    #         values = list(eval(el.get('limits')))  # make sure the evaluated values are returned as list (in case another
    #     # iterator type has been used
    #         param.update(dict(values=values))
    #     except:
    #         pass

    if 'limits' in el.attrib.keys():
        try:
            limits = eval(el.get('limits'))
            param.update(dict(limits=limits))
        except:
            pass

    return param


def parameter_to_xml_string(param):
    """ Convert  a Parameter to a XML string.

    Parameters
    ----------
    param: Parameter

    Returns
    -------
    str: XMl string

    See Also
    --------
    add_text_to_elt, walk_parameters_to_xml, dict_from_param

    Examples
    --------
    >>> from pyqtgraph.parametertree import Parameter
    >>>    #Create an instance of Parameter
    >>> settings=Parameter(name='settings')
    >>> converted_xml=parameter_to_xml_string(settings)
    >>>    # The converted Parameter
    >>> print(converted_xml)
    b'<settings title="settings" type="None" />'
    """
    xml_elt = walk_parameters_to_xml(param=param)
    return ET.tostring(xml_elt)


def parameter_to_xml_file(param, filename: Union[str, Path], overwrite=True):
    """
        Convert the given parameter to XML element and update the given XML file.

        =============== ================================= ================================
        **Parameters**    **Type**                          **Description**
        *param*           instance of pyqtgraph parameter   the parameter to be added
        *filename*        string                            the filename of the XML file
        *overwrite*       boolean                           raise Error is False and file exists
        =============== ================================= ================================

        See Also
        --------
        walk_parameters_to_xml

        Examples
        --------
    """
    if not isinstance(filename, Path):
        filename = Path(filename)
    parent = filename.parent
    filename = filename.stem
    fname = parent.joinpath(filename + ".xml")  # forcing the right extension on the filename
    xml_elt = walk_parameters_to_xml(param=param)
    tree = ET.ElementTree(xml_elt)
    if not overwrite and fname.exists() and fname.is_file():
        raise FileExistsError
    tree.write(str(fname))


def walk_xml_to_parameter(params=[], XML_elt=None):
    """ To convert an XML element (and children) to list of dict enabling creation of parameter object.

        =============== ================== =======================================
        **Parameters**   **Type**            **Description**

        *params*         dictionnary list    the list to create parameter object
        *XML_elt*        XML object          the XML object to be converted
        =============== ================== =======================================

        Returns
        -------
        params : dictionnary list
            list of dict to create parameter object

        Examples
        -------
        >>> from pyqtgraph.parametertree import Parameter, ParameterItem
        >>> import xml.etree.ElementTree as ET
        >>> tree = ET.parse('text_bis.xml')
        >>> root = tree.getroot()
        >>> params=walk_xml_to_parameter(XML_elt=root)
        >>> settings_xml=Parameter.create(name='Settings XML', type='group', children=params)
        >>> settings=Parameter.create(name='Settings', type='group', children=params)

        See Also
        --------
        walk_parameters_to_xml
    """
    try:
        if type(XML_elt) is not ET.Element:
            raise TypeError('not valid XML element')

        if len(XML_elt) == 0:
            param_dict = elt_to_dict(XML_elt)
            param_type = XML_elt.get('type')

            if 'group' not in param_type:  # covers 'group', custom 'groupmove'...
                set_txt_from_elt(XML_elt, param_dict)
            params.append(param_dict)

        for el in XML_elt:
            param_dict = elt_to_dict(el)
            param_type = el.get('type')

            if 'group' not in param_type:  # covers 'group', custom 'groupmove'...
                set_txt_from_elt(el, param_dict)
            else:
                if param_type not in PARAM_TYPES:
                    param_dict['type'] = 'group'  # in case the custom group has been defined somewhere but not
                    # registered again in this session
                if len(el) == 0:
                    children = []
                else:
                    subparams = []
                    children = walk_xml_to_parameter(subparams, el)
                param_dict['children'] = children

                param_dict['name'] = el.tag

            params.append(param_dict)
    except Exception as e:  # to be able to debug when there's an issue
        raise e
    return params


def set_txt_from_elt(el, param_dict):
    """
    get the value of the parameter from the text value of the xml element
    Parameters
    ----------
    el: xml element
    param_dict: dictionnary from which the parameter will be constructed

    """
    val_text = el.text
    param_type = el.get('type')
    if val_text is not None:
        if param_type == 'float':
            param_value = float(val_text)
        elif param_type == 'int':
            param_value = int(float(val_text))
        elif param_type == 'slide':
            param_value = float(val_text)
        elif param_type == 'itemselect':
            if val_text == 'None':
                param_value = dict(all_items=[], selected=[])
            else:
                param_value = dict(all_items=eval(el.get('all_items', val_text)), selected=eval(val_text))
        elif 'bool' in param_type or 'led' in param_type: # covers 'bool' 'bool_push',  'led' and 'led_push'types
            param_value = bool(int(val_text))
        elif param_type == 'date_time':
            param_value = QDateTime.fromMSecsSinceEpoch(int(val_text))
        elif param_type == 'date':
            param_value = QDateTime.fromMSecsSinceEpoch(int(val_text)).date()
        elif param_type == 'table':
            param_value = eval(val_text)
        elif param_type == 'color':
            param_value = QtGui.QColor(*eval(val_text))
        elif param_type == 'list':
            try:
                param_value = eval(val_text)
            except Exception:
                param_value = val_text  # for back compatibility
        elif param_type == 'table_view':
            data_dict = json.loads(val_text)
            mod = importlib.import_module(data_dict['module'])
            _cls = getattr(mod, data_dict['classname'])
            param_value = _cls(data_dict['data'], header=data_dict['header'])
        elif param_type == 'action':
            if val_text == 'None':
                param_value = None
            else:
                param_value = eval(val_text)
        else:
            param_value = val_text
        param_dict.update(dict(value=param_value))
    else:
        if param_type == 'str':
            param_dict.update(dict(value=''))


def XML_file_to_parameter(file_name: Union[str, Path]) -> list:
    """ Convert a xml file into pyqtgraph parameter object.

    Returns
    -------
    params : list of dictionary
        a list of dictionary defining a Parameter object and its children

    See Also
    --------
    walk_parameters_to_xml

    Examples
    --------
    """
    tree = ET.parse(str(file_name))

    root = tree.getroot()
    params = walk_xml_to_parameter(params=[], XML_elt=root)
    return params


def XML_string_to_parameter(xml_string):
    """
        Convert a xml string into a list of dict for initialize pyqtgraph parameter object.

        =============== =========== ================================
        **Parameters**   **Type**    **Description**

        xml_string       string      the xml string to be converted
        =============== =========== ================================

        Returns
        -------
        params: a parameter list of dict to init a parameter

        See Also
        --------
        walk_parameters_to_xml

        Examples
        --------
    """
    root = ET.fromstring(xml_string)
    tree = ET.ElementTree(root)

    # tree.write('test.xml')
    params = walk_xml_to_parameter(params=[], XML_elt=root)

    return params


def XML_string_to_pobject(xml_string) -> Parameter:
    """
    return a Parameter object from its *translated* version as a XML string
    Parameters
    ----------
    xml_string: (str) string representation of a Parameter Object

    Returns
    -------
    Parameter

    See Also
    --------
    parameter_to_xml_string
    """
    return Parameter.create(name='settings', type='group',
                            children=XML_string_to_parameter(xml_string))



from qtpy import QtGui, QtWidgets
from qtpy.QtCore import Slot, QObject, QThread, Signal, QLocale, QDateTime, QSize
import sys
import pyqtgraph as pg
import numpy as np

from pymodaq_utils.logger import set_logger, get_module_name
from pymodaq_utils import utils

logger = set_logger(get_module_name(__file__))


class Viewer1DBasic(QObject):
    """this plots 1D data on a plotwidget. one linear region to select data, one infinite line to select point
    """
    roi_region_signal = Signal(tuple)
    roi_line_signal = Signal(float)

    def __init__(self, parent=None, show_region=False, show_line=False):
        """

        Parameters
        ----------
        parent

        Attributes
        ----------
        parent: (QWidget)
        roi_region: (pyqtgraph LinerrRegionItem)
        roi_line: (pyqtgraph InfiniteLine graphitem)
        Properties
        ----------
        labels: (list of str)
        x_axis: (Axis or dict)

        """
        
        super().__init__()
        if parent is None:
            parent = QtWidgets.QWidget()
            parent.show()

        self.parent = parent

        self.plot_colors = utils.plot_colors

        self.data_to_export = None
        self.datas = None
        self._x_axis = None

        self.labels = []
        self.plot_channels = None
        self.legend = None
        self.setup_ui()

        self.show_roi_region(show_region)
        self.show_roi_line(show_line)

    def show(self, state=True):
        self.parent.setVisible(state)

    def setup_ui(self):
        vboxlayout = QtWidgets.QVBoxLayout()
        self.plotwidget = pg.PlotWidget()
        self.parent.setLayout(vboxlayout)
        vboxlayout.addWidget(self.plotwidget)
        vboxlayout.setContentsMargins(0, 0, 0, 0)

        self.legend = self.plotwidget.addLegend()
        self.roi_region = pg.LinearRegionItem()
        self.roi_line = pg.InfiniteLine(movable=True)
        self.plotwidget.plotItem.addItem(self.roi_region)
        self.plotwidget.plotItem.addItem(self.roi_line)
        self.roi_region.sigRegionChanged.connect(self.update_region)
        self.roi_line.sigPositionChanged.connect(self.update_line)

    def get_line_position(self):
        return self.roi_line.getXPos()

    def show_roi_region(self, show=True):
        self.roi_region.setVisible(show)

    def show_roi_line(self, show=True):
        self.roi_line.setVisible(show)

    def update_region(self, item):
        self.roi_region_signal.emit(item.getRegion())

    def update_line(self, item):
        self.roi_line_signal.emit(item.getPos()[0])

    @property
    def labels(self):
        return self._labels

    @labels.setter
    def labels(self, labels):
        self._labels = labels
        self.update_labels(labels)

    def update_labels(self, labels=[]):
        try:
            if self.datas is not None:
                labels_tmp = labels[:]
                if self.labels == labels:
                    if self.labels == [] or len(self.labels) < len(self.datas):
                        self._labels = ["CH{}".format(ind) for ind in range(len(self.datas))]
                else:
                    flag = True
                    while flag:
                        items = [item[1].text for item in self.legend.items]
                        if len(items) == 0:
                            flag = False
                        else:
                            self.legend.removeItem(items[0])

                    if len(labels) < len(self.plot_channels):
                        for ind in range(len(labels), len(self.plot_channels)):
                            labels_tmp.append('CH{:02d}'.format(ind))

                    if len(labels_tmp) == len(self.plot_channels):
                        for ind, channel in enumerate(self.plot_channels):
                            self.legend.addItem(channel, labels_tmp[ind])

                    self._labels = labels_tmp

        except Exception as e:
            logger.exception(str(e))

    @Slot(list)
    def show_data(self, datas):
        if datas is not None and datas[0] is not None:
            self.datas = datas

            if self.labels == [] or len(self.labels) != len(datas):
                self.update_labels(self.labels)

            if self.plot_channels is None:  # initialize data and plots
                self.ini_data_plots(len(datas))
            elif len(self.plot_channels) != len(datas):
                self.remove_plots()
                self.ini_data_plots(len(datas))

            for ind_plot, data in enumerate(datas):
                if self._x_axis is None or len(self._x_axis) != len(data):
                    self._x_axis = np.linspace(0, len(data), len(data), endpoint=False)

                self.plot_channels[ind_plot].setData(x=self.x_axis, y=data)

    def ini_data_plots(self, Nplots):
        self.plot_channels = []

        channels = []
        for ind in range(Nplots):
            channel = self.plotwidget.plot()
            channel.setPen(self.plot_colors[ind])
            self.legend.addItem(channel, self._labels[ind])
            channels.append(ind)
            self.plot_channels.append(channel)

    def remove_plots(self):
        if self.plot_channels is not None:
            for channel in self.plot_channels:
                self.plotwidget.removeItem(channel)
            self.plot_channels = None
        if self.legend is not None:
            items = [item[1].text for item in self.legend.items]
            for item in items:
                self.legend.removeItem(item)

    def set_axis_label(self, axis_settings=dict(orientation='bottom', label='x axis', units='pxls')):
        axis = self.plotwidget.plotItem.getAxis(axis_settings['orientation'])
        axis.setLabel(text=axis_settings['label'], units=axis_settings['units'])

    @property
    def x_axis(self):
        return self._x_axis

    @x_axis.setter
    def x_axis(self, x_axis):
        label = 'Pxls'
        units = ''
        if isinstance(x_axis, dict):
            if 'data' in x_axis:
                xdata = x_axis['data']
            if 'label' in x_axis:
                label = x_axis['label']
            if 'units' in x_axis:
                units = x_axis['units']
        else:
            xdata = x_axis
        self._x_axis = xdata
        self.show_data(self.datas)
        self.set_axis_label(dict(orientation='bottom', label=label, units=units))


if __name__ == '__main__':  # pragma: no cover
    app = QtWidgets.QApplication(sys.argv)

    def print_region(xx):
        print(xx)

    def print_line(x):
        print(x)

    Form = QtWidgets.QWidget()
    prog = Viewer1DBasic(Form)
    # prog.show_roi_region()
    prog.show_roi_line()
    prog.roi_region_signal.connect(print_region)
    prog.roi_line_signal.connect(print_line)
    from pymodaq.utils.daq_utils import gauss1D

    x = np.linspace(0, 200, 201)
    y1 = gauss1D(x, 75, 25)
    y2 = gauss1D(x, 120, 50, 2)
    tau_half = 27
    tau2 = 100
    x0 = 50
    dx = 20
    ydata_expodec = np.zeros((len(x)))
    ydata_expodec[:50] = 1 * gauss1D(x[:50], x0, dx, 2)
    ydata_expodec[50:] = 1 * np.exp(-(x[50:] - x0) / (tau_half / np.log(2)))  # +1*np.exp(-(x[50:]-x0)/tau2)
    ydata_expodec += 0.1 * np.random.rand(len(x))

    prog.show_data([y1, y2, ydata_expodec])

    Form.show()
    prog.x_axis
    prog.update_labels(labels=['sig0', 'tralala', 'ouhaouh'])
    sys.exit(app.exec_())



import sys
from qtpy import QtWidgets
from qtpy.QtCore import QDateTime, Signal, QPointF, QObject
import numpy as np
import pyqtgraph as pg
from pymodaq.utils.plotting.widgets.image import ImageWidget
from pymodaq.utils.managers.roi_manager import ROIBrushable

from pyqtgraph.graphicsItems.GradientEditorItem import Gradients

pos = []
colors = []
for ticks in Gradients['thermal']['ticks']:
    pos.append(ticks[0])
    colors.append(ticks[1])
cmap = pg.ColorMap(pos, colors, Gradients['thermal']['mode'])


def setTicksLabels(values):
    strings = []
    for v in values:
        d = QDateTime()
        d = d.fromSecsSinceEpoch(v * 30 * 60)
        vstr = d.toString('HH:mm dd/MM/yy')
        strings.append(vstr)
    return strings


class AxisItemDate(pg.AxisItem):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def tickStrings(self, values, scale, spacing):
        if self.logMode:
            return self.logTickStrings(values, scale, spacing)

        places = max(0, np.ceil(-np.log10(spacing * scale)))
        strings = setTicksLabels(values)
        return strings


class GanttROI(ROIBrushable):
    index_signal = Signal(int)

    def __init__(self, task='No task', index=0, start=0, stop=1, brush=None, *args, **kwargs):
        super().__init__(pos=[start, index], size=[stop - start, 1], brush=brush, pen=brush, snapSize=1, scaleSnap=True,
                         translateSnap=True, *args, **kwargs)  # )
        self.h1 = self.addScaleHandle([1, 0.5], [0, 0.5])
        self.h2 = self.addScaleHandle([0, 0.5], [1, 0.5])

        self.task = task
        self.setToolTip(task)
        self.index = index
        self.sigRegionChangeFinished.connect(self.emit_index_signal)
        self.sigRegionChangeFinished.connect(self.update_tooltip)
        self.update_tooltip()

    def update_tooltip(self):
        self.h2.setToolTip(setTicksLabels([self.pos()[0]])[0])
        self.h1.setToolTip(setTicksLabels([self.pos()[0] + self.size()[0]])[0])

    def center(self):
        return QPointF(self.pos().x() + self.size().x() / 2, self.pos().y() + self.size().y() / 2)

    def emit_index_signal(self):
        self.index_signal.emit(self.index)


class GanttTask(QObject):
    alpha = 255

    def __init__(self, task_dict):
        super().__init__()

        self.task_dict = task_dict  # taskdict=dict(name=  ,idnumber=  ,task_type=int(0 to 12), time_start=, time_end=)
        qc = cmap.mapToQColor(task_dict['task_type'] / 12)
        qc.setAlpha(self.alpha)
        self.roi_item = GanttROI(task_dict['name'], task_dict['idnumber'], start=task_dict['time_start'],
                                 stop=task_dict['time_end'], brush=qc)
        self.roi_item.sigRegionChangeFinished.connect(self.show_hide_move_text())
        self.roi_item.setZValue(-1000)
        self.text_item = pg.TextItem(task_dict['name'])
        self.text_item.setPos(task_dict['time_start'], task_dict['idnumber'])
        self.roi_item.setZValue(-500)

    def show_hide_move_text(self):
        pass


class GanttChart(QObject):
    def __init__(self):
        super().__init__()
        self.setupUI()
        self.tasks = []

    def setupUI(self):
        axis = AxisItemDate(orientation='bottom')
        axis_top = AxisItemDate(orientation='top')

        self.widget = ImageWidget(axisItems=dict(bottom=axis, top=axis_top))
        self.widget.plotitem.showAxis('top')
        self.widget.plotitem.vb.setAspectLocked(lock=False)
        self.widget.view.invertY(True)
        self.widget.view.setLimits(yMin=0)

    def add_task(self, task_dict):
        self.tasks.append(GanttTask(task_dict))
        self.widget.plotItem.addItem(self.tasks[-1].roi_item)
        self.widget.plotItem.addItem(self.tasks[-1].text_item)


# # create GUI
if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    gchart = GanttChart()
    gchart.widget.show()

    gchart.add_task(
        dict(name='This is task 1, the best of all the tasks!yeeah baby', idnumber=1, task_type=1, time_start=5,
             time_end=7))

    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
Created the 15/11/2022

@author: Sebastien Weber
"""
import copy
import datetime
from dateutil import parser
from numbers import Number
import os
from pathlib import Path
from typing import Union, Iterable


import numpy as np
from qtpy.QtCore import QObject, Signal
from qtpy import QtWidgets

from pymodaq_utils.logger import set_logger, get_module_name
from pymodaq_utils import utils
from pymodaq_utils.config import Config

from pymodaq_data.h5modules.backends import (
    H5Backend, backends_available, SaveType,
    GroupType, InvalidDataDimension, InvalidScanType,
    GROUP, VLARRAY)
from pymodaq_data.h5modules.saving import H5SaverLowLevel

from pymodaq_gui.parameter import Parameter, ParameterTree
from pymodaq_gui.parameter import utils as putils
from pymodaq_gui.managers.parameter_manager import ParameterManager
from pymodaq_gui.utils.file_io import select_file
from pymodaq_gui.h5modules import browsing

config = Config()
logger = set_logger(get_module_name(__file__))


dashboard_submodules_params = [
    {'title': 'Save 2D datas and above:', 'name': 'save_2D', 'type': 'bool', 'value': True},
    {'title': 'Save raw datas only:', 'name': 'save_raw_only', 'type': 'bool', 'value': True, 'tooltip':
        'if True, will not save extracted ROIs used to do live plotting, only raw datas will be saved'},
    {'title': 'Do Save:', 'name': 'do_save', 'type': 'bool', 'default': False, 'value': False},
    {'title': 'N saved:', 'name': 'N_saved', 'type': 'int', 'default': 0, 'value': 0, 'visible': False},
]


class H5SaverBase(H5SaverLowLevel, ParameterManager):
    """Object containing all methods in order to save datas in a *hdf5 file* with a hierarchy
    compatible with the H5Browser. The saving parameters are contained within a **Parameter**
    object: self.settings that can be displayed on a UI using the widget self.settings_tree.
    At the creation of a new file, a node group named **Raw_data** and represented by the attribute
    ``raw_group`` is created and set with a metadata attribute:

    * 'type' given by the **save_type** class parameter

    The root group of the file is then set with a few metadata:

    * 'pymodaq_version' the current pymodaq version, e.g. 1.6.2
    * 'pymodaq_data_version' the current pymodaq_data version, e.g. 0.0.1
    * 'file' the file name
    * 'date' the current date
    * 'time' the current time

    All data will then be saved under this node in various groups

    See Also
    --------
    H5Browser

    Parameters
    ----------
    h5_file: pytables hdf5 file
             object used to save all datas and metadas
    h5_file_path: str or Path
                  Signal signal represented by a float. Is emitted each time the hardware reached the target
                  position within the epsilon precision (see comon_parameters variable)
    save_type: str
       an element of the enum module attribute SaveType
       * 'scan' is used for DAQScan module and should be used for similar application
       * 'detector' is used for DAQ_Viewer module and should be used for similar application
       * 'custom' should be used for customized applications

    Attributes
    ----------

    settings: Parameter
               Parameter instance (pyqtgraph) containing all settings (could be represented using the settings_tree widget)

    settings_tree: ParameterTree
                   Widget representing as a Tree structure, all the settings defined in the class preamble variable ``params``

    """
    settings_name = 'h5saver_settings'
    params = [
        {'title': 'Save type:', 'name': 'save_type', 'type': 'list', 'limits': SaveType.names(),
         'readonly': True},
    ] + dashboard_submodules_params + \
        [
            {'title': 'Backend:', 'name': 'backend', 'type': 'group', 'children': [
            {'title': 'Backend type:', 'name': 'backend_type', 'type': 'list',
             'limits': backends_available, 'readonly': True},
            {'title': 'HSDS Server:', 'name': 'hsds_options', 'type': 'group', 'visible': False,
             'children': [
                {'title': 'Endpoint:', 'name': 'endpoint', 'type': 'str',
                    'value': config('data_saving', 'hsds', 'root_url'), 'readonly': False},
                {'title': 'User:', 'name': 'user', 'type': 'str',
                    'value': config('data_saving', 'hsds', 'username'), 'readonly': False},
                {'title': 'password:', 'name': 'password', 'type': 'str',
                    'value': config('data_saving', 'hsds', 'pwd'), 'readonly': False},
            ]},
        ]},
        {'title': 'custom_name?:', 'name': 'custom_name', 'type': 'bool', 'default': False,
         'value': False},
        {'title': 'show file content?', 'name': 'show_file', 'type': 'bool_push', 'default': False,
            'value': False},
        {'title': 'Base path:', 'name': 'base_path', 'type': 'browsepath',
            'value': config('data_saving', 'h5file', 'save_path'), 'filetype': False,
         'readonly': True, },
        {'title': 'Base name:', 'name': 'base_name', 'type': 'str', 'value': 'Scan',
         'readonly': True},
        {'title': 'Current scan:', 'name': 'current_scan_name', 'type': 'str', 'value': '',
         'readonly': True},
        {'title': 'Current path:', 'name': 'current_scan_path', 'type': 'text',
            'value': config('data_saving', 'h5file', 'save_path'), 'readonly': True,
         'visible': False},
        {'title': 'h5file:', 'name': 'current_h5_file', 'type': 'text', 'value': '',
         'readonly': True},
        {'title': 'New file', 'name': 'new_file', 'type': 'action'},
        {'title': 'Saving dynamic', 'name': 'dynamic', 'type': 'list',
         'limits': config('data_saving', 'data_type', 'dynamics'),
         'value': config('data_saving', 'data_type', 'dynamic')},
        {'title': 'Compression options:', 'name': 'compression_options', 'type': 'group',
         'children': [
            {'title': 'Compression library:', 'name': 'h5comp_library', 'type': 'list',
             'value': 'zlib', 'limits': ['zlib', 'gzip']},
            {'title': 'Compression level:', 'name': 'h5comp_level', 'type': 'int',
                'value': config('data_saving', 'h5file', 'compression_level'), 'min': 0, 'max': 9},
        ]},
    ]

    def __init__(self, save_type='scan', backend='tables'):
        """

        Parameters
        ----------
        save_type (str): one of ['scan', 'detector', 'logger', 'custom']
        backend (str): either 'tables' for pytables backend, 'h5py' for h5py backends or 'h5pyd' for HSDS backend

        See Also
        --------
        https://github.com/HDFGroup/hsds
        """
        H5SaverLowLevel.__init__(self, save_type, backend)
        ParameterManager.__init__(self)

        self.current_scan_group = None
        self.current_scan_name = None

        self.settings.child('save_type').setValue(self.save_type.name)

    def show_settings(self, show=True):
        self.settings_tree.setVisible(show)

    def init_file(self, update_h5=False, custom_naming=False, addhoc_file_path=None,
                  metadata=dict([])):
        """Initializes a new h5 file.
        Could set the h5_file attributes as:

        * a file with a name following a template if ``custom_naming`` is ``False`` and ``addhoc_file_path`` is ``None``
        * a file within a name set using a file dialog popup if ``custom_naming`` is ``True``
        * a file with a custom name if ``addhoc_file_path`` is a ``Path`` object or a path string

        Parameters
        ----------
        update_h5: bool
                   create a new h5 file with name specified by other parameters
                   if false try to open an existing file and will append new data to it
        custom_naming: bool
                       if True, a selection file dialog opens to set a new file name
        addhoc_file_path: Path or str
                          supplied name by the user for the new file
        metadata: dict
                    dictionnary with pair of key, value that should be saved as attributes of the root group
        Returns
        -------
        update_h5: bool
                   True if new file has been created, False otherwise
        """
        datetime_now = datetime.datetime.now()
        if addhoc_file_path is None:
            if not os.path.isdir(self.settings['base_path']):
                os.mkdir(self.settings['base_path'])

            # set the filename and path
            base_name = self.settings['base_name']

            if not custom_naming:
                custom_naming = self.settings['custom_name']

            if not custom_naming:
                scan_type = self.settings['save_type'] == 'scan'
                scan_path, current_scan_name, save_path = self.update_file_paths(update_h5)
                self.current_scan_name = current_scan_name
                self.settings.child('current_scan_name').setValue(current_scan_name)
                self.settings.child('current_scan_path').setValue(str(scan_path))

                if not scan_type:
                    self.h5_file_path = save_path.parent  # will remove the dataset part used for DAQ_scan datas
                    self.h5_file_name = base_name + datetime_now.strftime('_%Y%m%d_%H_%M_%S.h5')
                else:
                    self.h5_file_name = save_path.name + ".h5"
                    self.h5_file_path = save_path.parent

            else:
                self.h5_file_name = select_file(start_path=base_name, save=True, ext='h5')
                self.h5_file_path = self.h5_file_name.parent

        else:
            if isinstance(addhoc_file_path, str):
                addhoc_file_path = Path(addhoc_file_path)
            self.h5_file_path = addhoc_file_path.parent
            self.h5_file_name = addhoc_file_path.name

        fullpathname = self.h5_file_path.joinpath(self.h5_file_name)
        self.settings.child('current_h5_file').setValue(str(fullpathname))

        super().init_file(fullpathname, new_file=update_h5, metadata=metadata)

        self.get_set_logger(self.raw_group)

        return update_h5

    def update_file_paths(self, update_h5=False):
        """

        Parameters
        ----------
        update_h5: bool
                   if True, will increment the file name and eventually the current scan index
                   if False, get the current scan index in the h5 file

        Returns
        -------
        scan_path: Path
        current_filename: str
        dataset_path: Path

        """

        try:
            # set the filename and path
            base_path = self.settings['base_path']
            base_name = self.settings['base_name']
            current_scan = self.settings['current_scan_name']
            scan_type = self.settings['save_type'] == 'scan'
            ind_dataset = None
            if current_scan == '' or update_h5:
                next_scan_index = 0
                update_h5 = True  # just started the main program so one should create a new h5
                self.file_loaded = False
            else:
                next_scan_index = self.get_scan_index()
            if self.file_loaded:
                ind_dataset = int(os.path.splitext(self.h5_file_name)[0][-3:])
                try:
                    curr_date = datetime.date.fromisoformat(self.get_attr(self.root(), 'date'))
                except ValueError:
                    curr_date = parser.parse(self.get_attr(self.root(), 'date')).date()
            else:
                curr_date = datetime.date.today()

            scan_path, current_filename, dataset_path = self.set_current_scan_path(base_path, base_name, update_h5,
                                                                                   next_scan_index,
                                                                                   create_dataset_folder=False,
                                                                                   curr_date=curr_date,
                                                                                   ind_dataset=ind_dataset)
            self.settings.child('current_scan_path').setValue(str(dataset_path))

            return scan_path, current_filename, dataset_path

        except Exception as e:
            logger.exception(str(e))

    @classmethod
    def find_part_in_path_and_subpath(cls, base_dir, part='', create=False, increment=True):
        """
        Find path from part time.

        =============== ============ =============================================
        **Parameters**  **Type**      **Description**
        *base_dir*      Path object   The directory to browse
        *part*          string        The date of the directory to find/create
        *create*        boolean       Indicate the creation flag of the directory
        =============== ============ =============================================

        Returns
        -------
        Path object
            found path from part
        """
        found_path = None
        if part in base_dir.parts:  # check if current year is in the given base path
            if base_dir.name == part:
                found_path = base_dir
            else:
                for ind in range(len(base_dir.parts)):
                    tmp_path = base_dir.parents[ind]
                    if tmp_path.name == part:
                        found_path = base_dir.parents[ind]
                        break
        else:  # if not check if year is in the subfolders
            subfolders_year_name = [x.name for x in base_dir.iterdir() if x.is_dir()]
            subfolders_found_path = [x for x in base_dir.iterdir() if x.is_dir()]
            if part not in subfolders_year_name:
                if increment:
                    found_path = base_dir.joinpath(part)
                else:
                    found_path = base_dir
                if create:
                    found_path.mkdir()
            else:
                ind_path = subfolders_year_name.index(part)
                found_path = subfolders_found_path[ind_path]
        return found_path

    @classmethod
    def set_current_scan_path(cls, base_dir, base_name='Scan', update_h5=False, next_scan_index=0,
                              create_scan_folder=False,
                              create_dataset_folder=True, curr_date=None, ind_dataset=None):
        """

        Parameters
        ----------
        base_dir
        base_name
        update_h5
        next_scan_index
        create_scan_folder
        create_dataset_folder

        Returns
        -------

        """
        base_dir = Path(base_dir)
        if curr_date is None:
            curr_date = datetime.date.today()

        year_path = cls.find_part_in_path_and_subpath(base_dir, part=str(curr_date.year),
                                                      create=True)  # create directory of the year if it doen't exist and return it
        day_path = cls.find_part_in_path_and_subpath(year_path, part=curr_date.strftime('%Y%m%d'),
                                                     create=True)  # create directory of the day if it doen't exist and return it
        dataset_base_name = curr_date.strftime('Dataset_%Y%m%d')
        dataset_paths = sorted([path for path in day_path.glob(dataset_base_name + "*"+".h5") if path.is_file()])

        if ind_dataset is None:
            if dataset_paths == []:

                ind_dataset = 0
            else:
                if update_h5:
                    ind_dataset = int(dataset_paths[-1].stem.partition(dataset_base_name + "_")[2]) + 1
                else:
                    ind_dataset = int(dataset_paths[-1].stem.partition(dataset_base_name + "_")[2])

        dataset_path = cls.find_part_in_path_and_subpath(day_path,
                                                         part=dataset_base_name + "_{:03d}".format(ind_dataset),
                                                         create=False, increment=True)
        scan_paths = sorted([path for path in dataset_path.glob(base_name + '*') if path.is_dir()])
        ind_scan = next_scan_index
        return dataset_path, base_name + '{:03d}'.format(ind_scan), dataset_path

    def get_last_scan(self):
        """Gets the last scan node within the h5_file and under the **raw_group**

        Returns
        -------
        scan_group: pytables group or None


        """
        return self.get_last_group(self.raw_group, GroupType['scan'])

    def get_scan_groups(self):
        return self.get_groups(self.raw_group, GroupType['scan'])

    def get_scan_index(self):
        """ return the scan group index in the "scan templating": Scan000, Scan001 as an integer
        """

        last_scan = self.get_last_scan()
        return int(last_scan.name[4:]) if last_scan is not None else 0

    def load_file(self, base_path=None, file_path=None):
        """Opens a file dialog to select a h5file saved on disk to be used

        Parameters
        ----------
        base_path
        file_path

        See Also
        --------
        :py:meth:`init_file`

        """
        if base_path is None:
            base_path = self.settings.child('base_path').value()
            if not os.path.isdir(base_path):
                base_path = None

        if file_path is None:
            file_path = select_file(base_path, save=False, ext='h5')

        if not (file_path is None or file_path == ''):
            if not isinstance(file_path, Path):
                file_path = Path(file_path)

            if 'h5' not in file_path.suffix:
                raise IOError('Invalid file type, should be a h5 file')

            self.init_file(addhoc_file_path=file_path)
            self.file_loaded = True

    def save_file(self, filename=None):
        if filename is None:
            filename = select_file(None, save=True, ext='h5')
        if filename != '':
            super().save_file_as(filename)

    def value_changed(self, param):
        if param.name() == 'show_file':
            if param.value():
                param.setValue(False)
                self.show_file_content()

        elif param.name() == 'base_path':
            try:
                if not os.path.isdir(param.value()):
                    os.mkdir(param.value())
            except Exception as e:
                self.update_status(f"The base path couldn't be set, please check your options: {str(e)}")

        elif param.name() in putils.iter_children(self.settings.child('compression_options'), []):
            compression = self.settings.child('compression_options', 'h5comp_library').value()
            compression_opts = self.settings.child('compression_options', 'h5comp_level').value()
            self.define_compression(compression, compression_opts)

    def update_status(self, status):
        logger.warning(status)


class H5Saver(H5SaverBase, QObject):
    """
    status_sig: Signal
                emits a signal of type Threadcommand in order to senf log information to a main UI
    new_file_sig: Signal
                  emits a boolean signal to let the program know when the user pressed the new file button on the UI
    """

    status_sig = Signal(utils.ThreadCommand)
    new_file_sig = Signal(bool)

    def __init__(self, *args, **kwargs):
        """

        Parameters
        ----------
        args
        kwargs
        """
        QObject.__init__(self)
        H5SaverBase.__init__(self, *args, **kwargs)

        self.settings.child('new_file').sigActivated.connect(lambda: self.emit_new_file(True))

    def close(self):
        self.close_file()

    def emit_new_file(self, status):
        """Emits the new_file_sig

        Parameters
        ----------
        status: bool
                emits True if a new file has been asked by the user pressing the new file button on the UI
        """
        self.new_file_sig.emit(status)

    def show_file_content(self):
        win = QtWidgets.QMainWindow()
        if not self.isopen():
            if self.h5_file_path is not None:
                if self.h5_file_path.exists():
                    self.analysis_prog = browsing.H5Browser(win, h5file_path=self.h5_file_path)
                else:
                    logger.warning('The h5 file path has not been defined yet')
            else:
                logger.warning('The h5 file path has not been defined yet')
        else:
            self.flush()
            self.analysis_prog = browsing.H5Browser(win, h5file=self.h5file)
        win.show()


from pathlib import Path

from qtpy import QtWidgets, QtGui, QtCore
from qtpy.QtCore import Qt


class MySplash(QtWidgets.QSplashScreen):

    def __init__(self, *args, **kwargs):
        here = Path(__file__)
        pixmap = QtGui.QPixmap(str(here.parent.parent.joinpath('splash.png')))
        super().__init__(pixmap, Qt.WindowStaysOnTopHint, *args, **kwargs)

        font = self.font()
        font.setPixelSize(18)
        self.setFont(font)

    def showMessage(self, message, *args, **kwargs):
        """ force any message to be printed in white in the right/top corner """
        super().showMessage(message, QtCore.Qt.AlignRight, QtCore.Qt.white,
                            )

def get_splash_sc() -> MySplash:
    return MySplash()



from qtpy.QtCore import Qt
from qtpy import QtWidgets

from typing import Optional

from pymodaq_data.data import DataRaw, Axis

from pymodaq_gui.utils.widgets.table import SpinBoxDelegate
from pymodaq_gui.parameter.utils import get_widget_from_tree
from pymodaq_gui.plotting.data_viewers.viewer2D import Viewer2D
from pymodaq_gui.utils.custom_app import CustomApp, Dock, DockArea

from pymodaq_gui.parameter.pymodaq_ptypes.tableview import TableViewCustom
from pymodaq.utils.scanner.scanners.tabular import TableModelTabular


class DataPicker(CustomApp):
    params = [
        {'title': 'picker status', 'name': 'picker_status', 'type': 'led', 'value': False},
        {'title': 'Positions', 'name': 'tabular_table', 'type': 'table_view',
         'delegate': SpinBoxDelegate, 'menu': True}, ]

    def __init__(self, area: DockArea):
        super().__init__(area)

        self.viewer: Optional[Viewer2D] = None
        self.table_model: Optional[TableModelTabular] = None
        self.table_view: Optional[TableViewCustom] = None

        self.setup_ui()

    def setup_docks(self):
        self.docks['viewer'] = Dock('Viewer2D')
        self.dockarea.addDock(self.docks['viewer'], 'right')
        widget = QtWidgets.QWidget()
        self.viewer = Viewer2D(widget)
        self.docks['viewer'].addWidget(widget)

        self.docks['list'] = Dock('List of points')
        self.dockarea.addDock(self.docks['list'], 'right', self.docks['viewer'])
        self.docks['list'].addWidget(self.settings_tree)

        self.setup_table()

    def setup_actions(self):
        self.add_action('save_points', 'Save Points', 'move_contour',
                        tip='If checked, double clicking will put points in the table',
                        checkable=True)

    def setup_menu(self):
        action_menu = self._menubar.addMenu('Actions')
        self.affect_to('save_points', action_menu)

    def connect_things(self):
        self.connect_action('save_points',
                            lambda status: self.settings.child('picker_status').setValue(status))
        self.viewer.sig_double_clicked.connect(self.double_click_action)

    def double_click_action(self, posx: float, posy: float):
        if self.is_action_checked('save_points'):
            xs, ys = self.viewer.view.unscale_axis(posx, posy)
            data_at = self.viewer.view.get_data_at('red', (xs, ys))
            if data_at is not None:
                self.table_model.add_data(self.table_view.currentIndex().row() + 1,
                                          [posx, posy, data_at])

    def show_data(self, data: DataRaw):
        self.viewer.show_data(data)

    def setup_table(self):
        init_data = [[0., 0., 0.]]
        self.table_model = TableModelTabular(init_data, ['x', 'y', 'data'])
        self.table_view = get_widget_from_tree(self.settings_tree, TableViewCustom)[0]
        self.settings.child('tabular_table').setValue(self.table_model)

        self.table_view.horizontalHeader().ResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        self.table_view.setSelectionMode(QtWidgets.QTableView.SingleSelection)
        styledItemDelegate = QtWidgets.QStyledItemDelegate()
        styledItemDelegate.setItemEditorFactory(SpinBoxDelegate())
        self.table_view.setItemDelegate(styledItemDelegate)

        self.table_view.setDragEnabled(True)
        self.table_view.setDropIndicatorShown(True)
        self.table_view.setAcceptDrops(True)
        self.table_view.viewport().setAcceptDrops(True)
        self.table_view.setDefaultDropAction(Qt.MoveAction)
        self.table_view.setDragDropMode(QtWidgets.QTableView.InternalMove)
        self.table_view.setDragDropOverwriteMode(False)

        self.table_view.add_data_signal[int].connect(self.table_model.add_data)
        self.table_view.remove_row_signal[int].connect(self.table_model.remove_data)
        self.table_view.load_data_signal.connect(self.table_model.load_txt)
        self.table_view.save_data_signal.connect(self.table_model.save_txt)


def main():
    from pymodaq_gui.utils.utils import mkQApp
    import numpy as np

    app = mkQApp('DataPicker')
    area = DockArea()
    win = QtWidgets.QMainWindow()
    win.setCentralWidget(area)
    data_picker = DataPicker(area)

    Nx = 100
    Ny = 200
    x = (np.linspace(0, Nx - 1, Nx) + 100) / 2
    y = (np.linspace(0, Ny - 1, Ny) - 10) * 2
    from pymodaq.utils.math_utils import gauss2D

    data_red = 3 * gauss2D(x, np.mean(x), (np.max(x)-np.min(x)) / 5, y, np.mean(y), (np.max(y)-np.min(y)) / 5, 1)
    data_red += np.random.random(data_red.shape)
    data_to_plot = DataRaw(name='mydata', distribution='uniform', data=[data_red],
                           axes=[Axis('xaxis', units='m', data=x, index=1),
                                 Axis('yaxis', units='mm', data=y, index=0), ])

    data_picker.show_data(data_to_plot)
    win.show()
    app.exec()


if __name__ == '__main__':
    main()



from pyqtgraph.parametertree.parameterTypes.basetypes import SimpleParameter
from pyqtgraph.parametertree.parameterTypes.numeric import NumericParameterItem


class NumericParameter(SimpleParameter):
    itemClass = NumericParameterItem

    def __init__(self, **opts):
        super().__init__(**opts)

    def setLimits(self, limits):
        curVal = self.value()
        if curVal > limits[1]:
            self.setValue(limits[1])
        elif curVal < limits[0]:
            self.setValue(limits[0])
        super().setLimits(limits)
        return limits



import os
import sys
import warnings

from qtpy import QtCore
from qtpy.QtCore import QLocale
from pymodaq_utils import logger as logger_module
from pymodaq_utils.config import Config

from pymodaq_gui.qvariant import QVariant


logger = logger_module.set_logger(logger_module.get_module_name(__file__))

config = Config()


def set_qt_backend():
    backend_present = True
    if config('qtbackend', 'backend').lower() not in [mod.lower() for mod in sys.modules]:
        backend_present = False
        logger.warning(f"The chosen Qt backend ({config('qtbackend', 'backend')}) has not been installed...\n"
                       f"Trying another...")
        backends = config('qtbackend', 'backends')
        backends.pop(backends.index(config('qtbackend', 'backend')))
        for backend in backends:
            if backend.lower() in [mod.lower() for mod in sys.modules]:
                backend_present = True
                config['qtbackend', 'backend'] = backend
                break

    if backend_present:
        os.environ['QT_API'] = config('qtbackend', 'backend')
        logger.info('************************')
        logger.info(f"{config('qtbackend', 'backend')} Qt backend loaded")
        logger.info('************************')
    else:
        msg = f"No Qt backend could be found in your system, please install either pyqt5/6 or pyside2/6." \
              f"pyqt5 is still preferred, while pyqt6 should mostly work."
        logger.critical(msg)
        warnings.warn(msg, FutureWarning, 3)
        print(msg.upper())


def decode_data(encoded_data):
    """
    Decode QbyteArrayData generated when drop items in table/tree/list view
    Parameters
    ----------
    encoded_data: QByteArray
                    Encoded data of the mime data to be dropped
    Returns
    -------
    data: list
            list of dict whose key is the QtRole in the Model, and the value a QVariant

    """
    data = []

    ds = QtCore.QDataStream(encoded_data, QtCore.QIODevice.ReadOnly)
    while not ds.atEnd():
        row = ds.readInt32()
        col = ds.readInt32()

        map_items = ds.readInt32()
        item = {}
        for ind in range(map_items):
            key = ds.readInt32()
            #TODO check this is fine
            value = QVariant()
            #value = None
            ds >> value
            item[QtCore.Qt.ItemDataRole(key)] = value.value()
        data.append(item)
    return data


def setLocale():
    """
    defines the Locale to use to convert numbers to strings representation using language/country conventions
    Default is English and US
    """
    language = getattr(QLocale, config('style', 'language'))
    country = getattr(QLocale, config('style', 'country'))
    QLocale.setDefault(QLocale(language, country))




from typing import Union

from qtpy import QtCore, QtWidgets
from qtpy.QtCore import QObject, Slot, Signal
import sys
import pyqtgraph as pg
import numpy as np
from easydict import EasyDict as edict

from pymodaq_gui.plotting.widgets import ImageWidget
from pymodaq_gui.plotting.data_viewers.viewer import ViewerBase
from pymodaq_data.data import DataRaw
from pymodaq_gui.plotting.items.image import UniformImageItem, SpreadImageItem


class Viewer2DBasic(ViewerBase):
    """Very Basic 2D Viewer used as a view to add Image Items into it and use the image object directly

    Not meant to plot directly data, see Viewer2D for that

    See Also
    --------
    Navigator, UniformImageItem, SpreadImageItem

    """

    def __init__(self, parent: QtWidgets.QWidget = None, **kwargs):
        super().__init__(parent, **kwargs)

        self.scaling_options = edict(scaled_xaxis=edict(label="", units=None, offset=0, scaling=1),
                                     scaled_yaxis=edict(label="", units=None, offset=0, scaling=1))
        self.setup_ui()

    def scale_axis(self, xaxis, yaxis):
        return xaxis * self.scaling_options.scaled_xaxis.scaling + self.scaling_options.scaled_xaxis.offset, yaxis * self.scaling_options.scaled_yaxis.scaling + self.scaling_options.scaled_yaxis.offset

    @Slot(float, float)
    def double_clicked(self, posx, posy):
        self.sig_double_clicked.emit(posx, posy)

    def setup_ui(self):
        vlayout = QtWidgets.QVBoxLayout()
        hsplitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)

        self.parent.setLayout(vlayout)
        vlayout.addWidget(hsplitter)

        self.image_widget = ImageWidget()
        hsplitter.addWidget(self.image_widget)

        self.scaled_xaxis = self.image_widget.add_scaled_axis('top')
        self.scaled_yaxis = self.image_widget.add_scaled_axis('right')
        # self.scaled_xaxis.linkToView(self.image_widget.view)
        # self.scaled_yaxis.linkToView(self.image_widget.view)

        # self.image_widget.plotitem.layout.addItem(self.scaled_xaxis, *(1, 1))
        # self.image_widget.plotitem.layout.addItem(self.scaled_yaxis, *(2, 2))

        self.image_widget.view.sig_double_clicked.connect(self.double_clicked)


        # histograms
        self.histo_widget = QtWidgets.QWidget()
        histo_layout = QtWidgets.QHBoxLayout()
        self.histo_widget.setLayout(histo_layout)
        self.histogram_red = pg.HistogramLUTWidget()
        self.histogram_green = pg.HistogramLUTWidget()
        self.histogram_blue = pg.HistogramLUTWidget()
        self.histogram_adaptive = pg.HistogramLUTWidget()
        Ntick = 3
        colors_red = [(int(r), 0, 0) for r in np.linspace(0, 255, Ntick)]
        colors_green = [(0, int(g), 0) for g in np.linspace(0, 255, Ntick)]
        colors_blue = [(0, 0, int(b)) for b in np.linspace(0, 255, Ntick)]
        colors_adaptive = [(int(b), int(b), int(b)) for b in np.linspace(0, 255, Ntick)]
        cmap_red = pg.ColorMap(pos=np.linspace(0.0, 1.0, Ntick), color=colors_red)
        cmap_green = pg.ColorMap(pos=np.linspace(0.0, 1.0, Ntick), color=colors_green)
        cmap_blue = pg.ColorMap(pos=np.linspace(0.0, 1.0, Ntick), color=colors_blue)
        cmap_adaptive = pg.ColorMap(pos=np.linspace(0.0, 1.0, Ntick), color=colors_adaptive)

        self.histogram_red.gradient.setColorMap(cmap_red)
        self.histogram_green.gradient.setColorMap(cmap_green)
        self.histogram_blue.gradient.setColorMap(cmap_blue)
        self.histogram_adaptive.gradient.setColorMap(cmap_adaptive)

        histo_layout.addWidget(self.histogram_red)
        histo_layout.addWidget(self.histogram_green)
        histo_layout.addWidget(self.histogram_blue)
        histo_layout.addWidget(self.histogram_adaptive)
        hsplitter.addWidget(self.histo_widget)

    def _show_data(self, data: DataRaw):
        """This basic viewer is not meant to plot data directly

        """
        ...

    def set_aspect_ratio(self, status=True):
        self.image_widget.plotitem.vb.setAspectLocked(lock=status, ratio=1)

    @property
    def plotitem(self):
        return self.image_widget.plotitem

    @property
    def histograms(self):
        return [self.histogram_red, self.histogram_green, self.histogram_blue]

    def add_image_item(self, image: Union[SpreadImageItem, UniformImageItem], histogram: pg.HistogramLUTWidget = None):
        """Adds an image item to this viewer plotitem

        Links it to an existing histogram
        """
        self.plotitem.addItem(image)
        if histogram in self.histograms:
            histogram.item.setImageItem(image)


if __name__ == '__main__':  # pragma: no cover
    from pymodaq_gui.plotting.items.image import SpreadImageItem
    from qtpy import QtSvg, QtGui

    app = QtWidgets.QApplication(sys.argv)
    form = QtWidgets.QWidget()
    prog = Viewer2DBasic(form)
    img = SpreadImageItem()
    prog.image_widget.plotItem.addItem(img)

    svg_item = QtSvg.QGraphicsSvgItem()
    svg_renderer = QtSvg.QSvgRenderer(
        r'C:\Users\weber\Labo\Projet-Dossier candidature\Technical project\GDSII\wafer.svg')
    svg_item.setSharedRenderer(svg_renderer)

    form.show()

    data = np.load('../../../resources/triangulation_data.npy')
    img.setImage(data)

    prog.image_widget.plotitem.addItem(svg_item)
    curr_size = svg_renderer.defaultSize()
    real_size = svg_item.boundingRect()
    # tr = QtGui.QTransform()
    # tr.translate(rect.left(), rect.top())
    # tr.scale(300/rect.width(), 300/rect.height())
    # svg_item.setTransform(tr)
    pass
    sys.exit(app.exec_())




import os
import sys
from typing import List, TYPE_CHECKING, Tuple, Union

from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import QObject, Slot, Signal, QPointF
from qtpy.QtGui import QIcon, QPixmap
from collections import OrderedDict


from pymodaq_gui.parameter import utils as putils
from pymodaq_gui.parameter import ParameterTree, Parameter, ioxml, pymodaq_ptypes
from pyqtgraph.parametertree.parameterTypes.basetypes import GroupParameter

from pymodaq_gui.managers.action_manager import QAction

from pyqtgraph import ROI as pgROI

from pyqtgraph import functions as fn
from pyqtgraph import LinearRegionItem as pgLinearROI
from pymodaq_utils.utils import plot_colors
from pymodaq_utils.logger import get_module_name, set_logger
from pymodaq_gui.config import get_set_roi_path
from pymodaq_gui.utils import select_file
from pymodaq_gui.plotting.utils import plot_utils

import numpy as np
from pathlib import Path
from pymodaq_data.post_treatment.process_to_scalar import DataProcessorFactory

data_processors = DataProcessorFactory()

roi_path = get_set_roi_path()
logger = set_logger(get_module_name(__file__))


class ROIPositionMapper(QtWidgets.QWidget):
    """ Widget presenting a Tree structure representing a ROI positions.
    """

    def __init__(self, roi_pos, roi_size):
        super().__init__()
        self.roi_pos = roi_pos
        self.roi_size = roi_size

    def show_dialog(self):
        self.params = [
            {'name': 'position', 'type': 'group', 'children': [
                {'name': 'x0', 'type': 'float', 'value': self.roi_pos[0] + self.roi_size[0] / 2,
                 'step': 1},
                {'name': 'y0', 'type': 'float', 'value': self.roi_pos[1] + self.roi_size[1] / 2,
                 'step': 1}
            ]},
            {'name': 'size', 'type': 'group', 'children': [
                {'name': 'width', 'type': 'float', 'value': self.roi_size[0], 'step': 1},
                {'name': 'height', 'type': 'float', 'value': self.roi_size[1], 'step': 1}]
             }]

        dialog = QtWidgets.QDialog(self)
        vlayout = QtWidgets.QVBoxLayout()
        self.settings_tree = ParameterTree()
        vlayout.addWidget(self.settings_tree, 10)
        self.settings_tree.setMinimumWidth(300)
        self.settings = Parameter.create(name='settings', type='group', children=self.params)
        self.settings_tree.setParameters(self.settings, showTop=False)
        dialog.setLayout(vlayout)

        buttonBox = QtWidgets.QDialogButtonBox(parent=self)
        buttonBox.addButton('Apply', buttonBox.AcceptRole)
        buttonBox.accepted.connect(dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(dialog.reject)

        vlayout.addWidget(buttonBox)
        self.setWindowTitle('Set Precise positions for the ROI')
        res = dialog.exec()

        if res == dialog.Accepted:

            return self.settings
        else:
            return None


class ROI(pgROI):
    index_signal = Signal(int)

    def __init__(self, *args, index=0, name='roi', **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.index = index
        self._menu = QtWidgets.QMenu()
        self._menu.addAction('Set ROI positions', self.set_positions)
        self._menu.addAction('Copy ROI to clipboard', self.copy_clipboard)
        self.sigRegionChangeFinished.connect(self.emit_index_signal)
        self._clipboard = QtGui.QGuiApplication.clipboard()

    def emit_index_signal(self):
        self.index_signal.emit(self.index)

    @property
    def color(self):
        return self.pen.color()

    def center(self) -> QPointF:
        """ Get the center position of the ROI """
        return QPointF(self.pos().x() + self.size().x() / 2, self.pos().y() + self.size().y() / 2)

    def set_center(self, center: Union[QPointF, Tuple[float, float]]):
        size = self.size()
        self.setPos(np.array(center) - np.array(size) / 2)

    def set_positions(self):
        mapper = ROIPositionMapper(self.pos(), self.size())
        settings = mapper.show_dialog()
        if settings is not None:
            self.setSize((settings['size', 'width'], settings['size', 'height']))
            self.setPos((settings['position', 'x0'] - settings['size', 'width'] / 2,
                         settings['position', 'y0'] - settings['size', 'height'] / 2))

    def copy_clipboard(self):
        info = plot_utils.RoiInfo.info_from_rect_roi(self)
        self._clipboard.setText(str(info.to_slices()))

    def contextMenuEvent(self, event):
        if self._menu is not None:
            self._menu.exec(event.screenPos())

    def width(self) -> float:
        return self.size().x()

    def height(self) -> float:
        return self.size().y()


class ROIBrushable(ROI):
    def __init__(self, brush=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if brush is None:
            brush = QtGui.QBrush(QtGui.QColor(0, 0, 255, 50))
        self.setBrush(brush)

    def setBrush(self, *br, **kargs):
        """Set the brush that fills the region. Can have any arguments that are valid
        for :func:`mkBrush <pyqtgraph.mkBrush>`.
        """
        self.brush = fn.mkBrush(*br, **kargs)
        self.currentBrush = self.brush

    def paint(self, p, opt, widget):
        # p.save()
        # Note: don't use self.boundingRect here, because subclasses may need to redefine it.
        r = QtCore.QRectF(0, 0, self.state['size'][0], self.state['size'][1]).normalized()

        p.setRenderHint(QtGui.QPainter.Antialiasing)
        p.setPen(self.currentPen)
        p.setBrush(self.currentBrush)
        p.translate(r.left(), r.top())
        p.scale(r.width(), r.height())
        p.drawRect(0, 0, 1, 1)
        # p.restore()


class LinearROI(pgLinearROI):
    index_signal = Signal(int)

    def __init__(self, index=0, pos=[0, 10], name = 'roi', **kwargs):
        super().__init__(values=pos, **kwargs)
        self.name = name
        self.index = index
        self.sigRegionChangeFinished.connect(self.emit_index_signal)

        self._menu = QtWidgets.QMenu()
        self._menu.addAction('Copy ROI to clipboard', self.copy_clipboard)
        self.sigRegionChangeFinished.connect(self.emit_index_signal)
        self._clipboard = QtGui.QGuiApplication.clipboard()

    def copy_clipboard(self):
        info = plot_utils.RoiInfo.info_from_linear_roi(self)
        self._clipboard.setText(str(info.to_slices()))

    def contextMenuEvent(self, event):
        if self._menu is not None:
            self._menu.exec(event.screenPos())

    def pos(self) -> Tuple[float, float]:
        return self.getRegion()

    def center(self) -> float:
        pos = self.pos()
        return (pos[0] + pos[1]) / 2

    def setPos(self, pos: Tuple[int, int]):
        self.setRegion(pos)

    def setPen(self, color):
        self.setBrush(color)

    @property
    def color(self):
        return self.brush.color()

    def emit_index_signal(self):
        self.index_signal.emit(self.index)


class EllipseROI(ROI):
    """
    Elliptical ROI subclass with one scale handle and one rotation handle.


    ============== =============================================================
    **Arguments**
    pos            (length-2 sequence) The position of the ROI's origin.
    size           (length-2 sequence) The size of the ROI's bounding rectangle.
    **args         All extra keyword arguments are passed to ROI()
    ============== =============================================================

    """


    def __init__(self, index=0, pos=[0, 0], size=[10, 10], **kwargs):
        # QtGui.QGraphicsRectItem.__init__(self, 0, 0, size[0], size[1])
        super().__init__(pos=pos, size=size, index=index, **kwargs)
        self.addRotateHandle([1.0, 0.5], [0.5, 0.5])
        self.addScaleHandle([0.5 * 2. ** -0.5 + 0.5, 0.5 * 2. ** -0.5 + 0.5], [0.5, 0.5])

    def getArrayRegion(self, arr, img=None, axes=(0, 1), **kwds):
        """
        Return the result of ROI.getArrayRegion() masked by the elliptical shape
        of the ROI. Regions outside the ellipse are set to 0.
        """
        # Note: we could use the same method as used by PolyLineROI, but this
        # implementation produces a nicer mask.
        if kwds["returnMappedCoords"]:
            arr, coords = pgROI.getArrayRegion(self, arr, img, axes, **kwds)
        else:
            arr = pgROI.getArrayRegion(self, arr, img, axes, **kwds)
        if arr is None or arr.shape[axes[0]] == 0 or arr.shape[axes[1]] == 0:
            return arr
        w = arr.shape[axes[0]]
        h = arr.shape[axes[1]]
        # generate an ellipsoidal mask
        mask = np.fromfunction(
            lambda x, y: (((x + 0.5) / (w / 2.) - 1) ** 2 + ((y + 0.5) / (h / 2.) - 1) ** 2) ** 0.5 < 1, (w, h))

        # reshape to match array axes
        if axes[0] > axes[1]:
            mask = mask.T
        shape = [(n if i in axes else 1) for i, n in enumerate(arr.shape)]
        mask = mask.reshape(shape)
        if kwds["returnMappedCoords"]:
            return arr * mask, coords
        else:
            return arr * mask

    def paint(self, p, opt, widget):
        r = self.boundingRect()
        p.setRenderHint(QtGui.QPainter.Antialiasing)
        p.setPen(self.currentPen)

        p.scale(r.width(), r.height())  # workaround for GL bug
        r = QtCore.QRectF(r.x() / r.width(), r.y() / r.height(), 1, 1)

        p.drawEllipse(r)

    def shape(self):
        self.path = QtGui.QPainterPath()
        self.path.addEllipse(self.boundingRect())
        return self.path


class CircularROI(EllipseROI):
    def __init__(self, index=0, pos=[0, 0], size=[10, 10], **kwargs):
        ROI.__init__(self, pos=pos, size=size, index=index, **kwargs)
        self.addScaleHandle([0.5 * 2. ** -0.5 + 0.5, 0.5 * 2. ** -0.5 + 0.5], [0.5, 0.5],
                            lockAspect=True)


class SimpleRectROI(ROI):
    r"""
    Rectangular ROI subclass with a single scale handle at the top-right corner.
    """

    def __init__(self, pos=[0, 0], size=[10, 10], centered=False, sideScalers=False, **args):
        super().__init__(pos, size, **args)
        if centered:
            center = [0.5, 0.5]
        else:
            center = [0, 0]

        self.addScaleHandle([1, 1], center)
        if sideScalers:
            self.addScaleHandle([1, 0.5], [center[0], 0.5])
            self.addScaleHandle([0.5, 1], [0.5, center[1]])


class RectROI(ROI):
    def __init__(self, index=0, pos=[0, 0], size=[10, 10], **kwargs):
        super().__init__(pos=pos, size=size, index=index, **kwargs)  # , scaleSnap=True, translateSnap=True)
        self.addScaleHandle([1, 1], [0, 0])
        self.addRotateHandle([0, 0], [0.5, 0.5])


ROI_NAME_PREFIX = 'ROI_'
ROI2D_TYPES = ['RectROI', 'EllipseROI', 'CircularROI']


class ROIScalableGroup(GroupParameter):
    def __init__(self, roi_type='1D', **opts):
        opts['type'] = 'group'
        opts['addText'] = "Add"
        self.roi_type = roi_type
        if roi_type != '1D':
            opts['addList'] = ROI2D_TYPES
        self.color_list = ROIManager.color_list
        super().__init__(**opts)

    def addNew(self, typ=''):
        name_prefix = ROI_NAME_PREFIX
        child_indexes = [int(par.name()[len(name_prefix) + 1:]) for par in self.children()]
        if not child_indexes:
            newindex = 0
        else:
            newindex = max(child_indexes) + 1

        child = {'name': ROIManager.roi_format(newindex), 'type': 'group', 'removable': True, 'renamable': False}

        children = [{'name': 'type', 'type': 'str', 'value': self.roi_type, 'readonly': True, 'visible': False}, ]
        if self.roi_type == '2D':
            children.extend([{'title': 'ROI Type', 'name': 'roi_type', 'type': 'str', 'value': typ, 'readonly': True},
                             {'title': 'Use channel', 'name': 'use_channel', 'type': 'list',
                              'limits': ['red', 'green', 'blue']}, ])
            children.append({'title': 'Math type:', 'name': 'math_function', 'type': 'list',
                             'limits': data_processors.functions_filtered('Data2D')})
        else:
            children.append({'title': 'Use channel', 'name': 'use_channel', 'type': 'list'})
            children.append({'title': 'Math type:', 'name': 'math_function', 'type': 'list',
                             'limits': data_processors.functions_filtered('Data1D')})

        children.extend([
            {'name': 'Color', 'type': 'color', 'value': list(np.roll(self.color_list, newindex)[0])}, ])
        if self.roi_type == '2D':
            children.extend([{'name': 'position', 'type': 'group', 'children': [
                {'name': 'x', 'type': 'float', 'value': 0, 'step': 1},
                {'name': 'y', 'type': 'float', 'value': 0, 'step': 1}
            ]}, ])
        else:
            children.extend([{'name': 'position', 'type': 'group', 'children': [
                {'name': 'left', 'type': 'float', 'value': 0, 'step': 1},
                {'name': 'right', 'type': 'float', 'value': 10, 'step': 1}
            ]}, ])
        if self.roi_type == '2D':
            children.extend([
                {'name': 'size', 'type': 'group', 'children': [
                    {'name': 'width', 'type': 'float', 'value': 10, 'step': 1},
                    {'name': 'height', 'type': 'float', 'value': 10, 'step': 1}
                ]},
                {'name': 'angle', 'type': 'float', 'value': 0, 'step': 1}])

        child['children'] = children

        self.addChild(child)


class ROIManager(QObject):

    new_ROI_signal = Signal(int, str, str)
    remove_ROI_signal = Signal(str)
    roi_value_changed = Signal(str, tuple)
    color_signal = Signal(list)
    roi_update_children = Signal(list)
    roi_changed = Signal()
    color_list = np.array(plot_colors)

    def __init__(self, viewer_widget=None, ROI_type='1D'):
        super().__init__()
        self.ROI_type = ROI_type
        self.roiwidget = QtWidgets.QWidget()
        self.viewer_widget = viewer_widget  # either a PlotWidget or a ImageWidget
        self._ROIs: OrderedDict[str, ROI] = OrderedDict([])
        self.setupUI()

    @staticmethod
    def roi_format(index):
        return f'{ROI_NAME_PREFIX}{index:02d}'

    @property
    def ROIs(self):
        return self._ROIs

    def __len__(self):
        return len(self._ROIs)

    def get_roi_from_index(self, index: int) -> ROI:
        return self.ROIs[self.roi_format(index)]

    def _set_roi_from_index(self, index: int, roi):
        self.ROIs[self.roi_format(index)] = roi

    def get_roi(self, roi_key):
        if roi_key in self.ROIs:
            return self.ROIs[roi_key]
        else:
            raise KeyError(f'{roi_key} is not a valid ROI identifier for {self.ROIs}')

    def emit_colors(self):
        self.color_signal.emit([self._ROIs[roi_key].color for roi_key in self._ROIs])

    def add_roi_programmatically(self, roitype=ROI2D_TYPES[0]):
        self.settings.child('ROIs').addNew(roitype)

    def remove_roi_programmatically(self, index: int):
        self.settings.child('ROIs').removeChild(self.settings.child('ROIs', self.roi_format(index)))

    def setupUI(self):

        vlayout = QtWidgets.QVBoxLayout()
        self.roiwidget.setLayout(vlayout)

        self.toolbar = QtWidgets.QToolBar()
        vlayout.addWidget(self.toolbar)

        self.save_ROI_pb = QAction(QIcon(QPixmap(":/icons/Icon_Library/save_ROI.png")), 'Save ROIs')
        self.load_ROI_pb = QAction(QIcon(QPixmap(":/icons/Icon_Library/load_ROI.png")), 'Load ROIs')
        self.clear_ROI_pb = QAction(QIcon(QPixmap(":/icons/Icon_Library/clear_ROI.png")), 'Clear ROIs')
        self.toolbar.addActions([self.save_ROI_pb, self.load_ROI_pb, self.clear_ROI_pb])


        self.roitree = ParameterTree()
        vlayout.addWidget(self.roitree)
        self.roiwidget.setMinimumWidth(250)
        self.roiwidget.setMaximumWidth(300)

        params = [
            {'title': 'Measurements:', 'name': 'measurements', 'type': 'table', 'value': OrderedDict([]), 'Ncol': 2,
             'header': ["LO", "Value"]},
            ROIScalableGroup(roi_type=self.ROI_type, name="ROIs")]
        self.settings = Parameter.create(title='ROIs Settings', name='rois_settings', type='group', children=params)
        self.roitree.setParameters(self.settings, showTop=False)
        self.settings.sigTreeStateChanged.connect(self.roi_tree_changed)

        self.save_ROI_pb.triggered.connect(self.save_ROI)
        self.load_ROI_pb.triggered.connect(lambda: self.load_ROI(None))
        self.clear_ROI_pb.triggered.connect(self.clear_ROI)

    def roi_tree_changed(self, param, changes):

        for param, change, data in changes:
            path = self.settings.childPath(param)
            if path is not None:
                childName = '.'.join(path)
            else:
                childName = param.name()
            if change == 'childAdded':  # new roi to create
                par: Parameter = data[0]
                newindex = int(par.name()[-2:])
                roi_type = ''
                if par.child('type').value() == '1D':
                    roi_type = ''

                    pos = self.viewer_widget.plotItem.vb.viewRange()[0]
                    pos = pos[0] + np.diff(pos)*np.array([2,4])/6
                    newroi = LinearROI(index=newindex, pos=pos)

                    newroi.setZValue(-10)
                    newroi.setBrush(par.child('Color').value())
                    newroi.setOpacity(0.2)

                elif par.child('type').value() == '2D':
                    roi_type = par.child('roi_type').value()
                    xrange = self.viewer_widget.plotItem.vb.viewRange()[0]
                    yrange = self.viewer_widget.plotItem.vb.viewRange()[1]
                    width = np.max(((xrange[1] - xrange[0]) / 10, 2))
                    height = np.max(((yrange[1] - yrange[0]) / 10, 2))
                    pos = [int(np.mean(xrange) - width / 2), int(np.mean(yrange) - width / 2)]

                    if roi_type == 'RectROI':
                        newroi = RectROI(index=newindex, pos=pos,
                                         size=[width, height], name=par.name())
                    elif roi_type == 'EllipseROI':
                        newroi = EllipseROI(index=newindex, pos=pos,
                                            size=[width, height], name=par.name())
                    elif roi_type == 'CircularROI':
                        newroi = CircularROI(index=newindex, pos=pos,
                                             size=[width, height], name=par.name())
                    newroi.setPen(par['Color'])

                newroi.sigRegionChangeFinished.connect(lambda: self.roi_changed.emit())
                newroi.index_signal[int].connect(self.update_roi_tree)
                try:
                    self.settings.sigTreeStateChanged.disconnect()
                except Exception:
                    pass
                self.settings.sigTreeStateChanged.connect(self.roi_tree_changed)
                self.viewer_widget.plotItem.addItem(newroi)

                self._set_roi_from_index(newindex, newroi)

                self.new_ROI_signal.emit(newindex, roi_type, par.name())
                self.update_roi_tree(newindex)
                self.emit_colors()
                self.roi_changed.emit()

            elif change == 'value':
                if param.name() in putils.iter_children(self.settings.child('ROIs'), []):
                    parent_name = putils.get_param_path(param)[putils.get_param_path(param).index('ROIs')+1]
                    self.update_roi(parent_name, param)
                    self.roi_value_changed.emit(parent_name, (param, param.value()))
                if param.name() == 'Color':
                    self.emit_colors()

            elif change == 'parent':
                if 'ROI' in param.name():
                    roi = self._ROIs.pop(param.name())
                    self.viewer_widget.plotItem.removeItem(roi)
                    self.remove_ROI_signal.emit(param.name())
                    self.emit_colors()

    def update_use_channel(self, channels: List[str]):
        channels.append('All')
        for ind in range(len(self)):
            val = self.settings['ROIs', self.roi_format(ind), 'use_channel']
            self.settings.child('ROIs', self.roi_format(ind), 'use_channel').setLimits(channels)
            if val not in channels:
                self.settings.child('ROIs', self.roi_format(ind), 'use_channel').setValue(channels[0])

    def update_roi(self, roi_key, param):
        self._ROIs[roi_key].index_signal[int].disconnect()
        if param.name() == 'Color':
            self._ROIs[roi_key].setPen(param.value())
            self.emit_colors()
        elif param.name() == 'left' or param.name() == 'x':
            pos = self._ROIs[roi_key].pos()
            poss = [param.value(), pos[1]]
            if self.settings.child('ROIs', roi_key, 'type').value() == '1D':
                poss.sort()
            self._ROIs[roi_key].setPos(poss)

        elif param.name() == 'right' or param.name() == 'y':
            pos = self._ROIs[roi_key].pos()
            poss = [pos[0], param.value()]
            if self.settings.child('ROIs', roi_key, 'type').value() == '1D':
                poss.sort()
            self._ROIs[roi_key].setPos(poss)

        elif param.name() == 'angle':
            self._ROIs[roi_key].setAngle(param.value(),center=[0.5,0.5])
        elif param.name() == 'width':
            size = self._ROIs[roi_key].size()
            self._ROIs[roi_key].setSize((param.value(), size[1]))
        elif param.name() == 'height':
            size = self._ROIs[roi_key].size()
            self._ROIs[roi_key].setSize((size[0], param.value()))
        self._ROIs[roi_key].index_signal[int].connect(self.update_roi_tree)

    @Slot(int)
    def update_roi_tree(self, index):
        roi = self.get_roi_from_index(index)
        par = self.settings.child(*('ROIs', self.roi_format(index)))
        if isinstance(roi, LinearROI):
            pos = roi.getRegion()
        else:
            pos = roi.pos()
            size = roi.size()
            angle = roi.angle()

        try:
            self.settings.sigTreeStateChanged.disconnect()
        except Exception:
            pass
        if isinstance(roi, LinearROI):
            par.child(*('position', 'left')).setValue(pos[0])
            par.child(*('position', 'right')).setValue(pos[1])
        if not isinstance(roi, LinearROI):
            par.child(*('position', 'x')).setValue(pos[0])
            par.child(*('position', 'y')).setValue(pos[1])
            par.child(*('size', 'width')).setValue(size[0])
            par.child(*('size', 'height')).setValue(size[1])
            par.child('angle').setValue(angle)

        self.settings.sigTreeStateChanged.connect(self.roi_tree_changed)

    def save_ROI(self):

        try:
            data = ioxml.parameter_to_xml_string(self.settings.child(('ROIs')))
            path = select_file(start_path=Path.home(), ext='xml', save=True, force_save_extension=True)

            if path != '':
                with open(path, 'wb') as f:
                    f.write(data)
        except Exception as e:
            print(e)

    def clear_ROI(self):
        indexes = [roi.index for roi in self._ROIs.values()]
        for index in indexes:
            self.settings.child(*('ROIs', self.roi_format(index))).remove()
            # self.settings.sigTreeStateChanged.connect(self.roi_tree_changed)

    def load_ROI(self, path=None, params=None):
        try:
            if params is None:
                if path is None:
                    path = select_file(start_path=Path.home(), save=False, ext='xml', filter='XML files (*.xml)')
                    if path != '':
                        params = Parameter.create(title='Settings', name='settings', type='group',
                                                  children=ioxml.XML_file_to_parameter(path))

            if params is not None:
                self.clear_ROI()
                QtWidgets.QApplication.processEvents()

                for param in params:
                    if 'roi_type' in putils.iter_children(param, []):
                        self.settings.child('ROIs').addNew(param.child('roi_type').value())
                    else:
                        self.settings.child('ROIs').addNew()
                QtWidgets.QApplication.processEvents()
                self.set_roi(self.settings.child('ROIs').children(), params)
        except Exception as e:
            logger.exception(str(e))

    def set_roi(self, roi_params, roi_params_new):
        for child, new_child in zip(roi_params, roi_params_new):
            if 'group' not in child.opts['type']:
                child.setValue(new_child.value())
            else:
                self.set_roi(child.children(), new_child.children())


class ROISaver:
    def __init__(self, msgbox=False, det_modules=[]):

        self.roi_presets = None
        self.detector_modules = det_modules

        if msgbox:
            msgBox = QtWidgets.QMessageBox()
            msgBox.setText("ROI Manager?")
            msgBox.setInformativeText("What do you want to do?")
            cancel_button = msgBox.addButton(QtWidgets.QMessageBox.Cancel)
            modify_button = msgBox.addButton('Modify', QtWidgets.QMessageBox.AcceptRole)
            msgBox.setDefaultButton(QtWidgets.QMessageBox.Cancel)
            ret = msgBox.exec()

            if msgBox.clickedButton() == modify_button:
                path = select_file(start_path=roi_path, save=False, ext='xml')
                if path != '':
                    self.set_file_roi(str(path))
            else:  # cancel
                pass

    def set_file_roi(self, filename, show=True):
        """

        """

        children = ioxml.XML_file_to_parameter(filename)
        self.roi_presets = Parameter.create(title='roi', name='rois', type='group', children=children)

        det_children = [child for child in self.roi_presets.children() if 'det' in child.opts['name']]
        det_names = [child.child('detname').value() for child in self.roi_presets.children() if
                     'det' in child.opts['name']]
        det_module_names = [det.title for det in self.detector_modules]
        for ind_det, det_roi in enumerate(det_children):
            det_module = self.detector_modules[det_module_names.index(det_names[ind_det])]
            viewer_children = [child for child in det_roi.children() if 'viewer' in child.opts['name']]
            for ind_viewer, viewer in enumerate(det_module.viewers):
                rois_params = [child for child in viewer_children[ind_viewer].children() if 'ROI' in child.opts['name']]
                if len(rois_params) > 0:
                    if hasattr(viewer, 'roi_manager'):
                        if hasattr(viewer, 'activate_roi'):  # because for viewer 0D it is irrelevant
                            viewer.activate_roi()
                        viewer.roi_manager.load_ROI(params=rois_params)
                        QtWidgets.QApplication.processEvents()

        if show:
            self.show_rois()

    def set_new_roi(self, file=None):
        if file is None:
            file = 'roi_default'

        self.roi_presets = Parameter.create(name='roi_settings', type='group', children=[
            {'title': 'Filename:', 'name': 'filename', 'type': 'str', 'value': file}, ])

        for ind_det, det in enumerate(self.detector_modules):
            det_param = Parameter.create(name=f'det_{ind_det:03d}', type='group', children=[
                {'title': 'Det Name:', 'name': 'detname', 'type': 'str', 'value': det.title}, ])

            for ind_viewer, viewer in enumerate(det.ui.viewers):
                viewer_param = Parameter.create(
                    name=f'viewer_{ind_viewer:03d}', type='group',
                    children=[
                        {'title': 'Viewer:', 'name': 'viewername', 'type': 'str',
                         'value': det.ui.viewer_docks[ind_viewer].name()}, ])

                if hasattr(viewer, 'roi_manager'):
                    viewer_param.addChild(
                        {'title': 'ROI type:', 'name': 'roi_type', 'type': 'str',
                         'value': viewer.roi_manager.settings.child('ROIs').roi_type})
                    viewer_param.addChildren(viewer.roi_manager.settings.child('ROIs').children())
                det_param.addChild(viewer_param)
            self.roi_presets.addChild(det_param)

        ioxml.parameter_to_xml_file(self.roi_presets, os.path.join(roi_path, file))
        self.show_rois()

    def show_rois(self):
        """

        """
        dialog = QtWidgets.QDialog()
        vlayout = QtWidgets.QVBoxLayout()
        tree = ParameterTree()
        tree.setMinimumWidth(400)
        tree.setMinimumHeight(500)
        tree.setParameters(self.roi_presets, showTop=False)

        vlayout.addWidget(tree)
        dialog.setLayout(vlayout)
        buttonBox = QtWidgets.QDialogButtonBox(parent=dialog)

        buttonBox.addButton('Save', buttonBox.AcceptRole)
        buttonBox.accepted.connect(dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(dialog.reject)

        vlayout.addWidget(buttonBox)
        dialog.setWindowTitle('Fill in information about this manager')
        res = dialog.exec()

        if res == dialog.Accepted:
            # save managers parameters in a xml file
            # start = os.path.split(os.path.split(os.path.realpath(__file__))[0])[0]
            # start = os.path.join("..",'daq_scan')
            ioxml.parameter_to_xml_file(
                self.roi_presets, os.path.join(
                    roi_path, self.roi_presets.child('filename').value()))


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    from pymodaq_gui.plotting.widgets import ImageWidget
    from pyqtgraph import PlotWidget

    im = ImageWidget()
    im = PlotWidget()
    prog = ROIManager(im, '2D')
    widget = QtWidgets.QWidget()
    layout = QtWidgets.QHBoxLayout()
    widget.setLayout(layout)
    layout.addWidget(im)
    layout.addWidget(prog.roiwidget)
    widget.show()
    prog.add_roi_programmatically(ROI2D_TYPES[0])
    prog.add_roi_programmatically(ROI2D_TYPES[1])
    prog.add_roi_programmatically(ROI2D_TYPES[2])
    sys.exit(app.exec_())


import sys
import datetime
from collections import OrderedDict
from typing import List, Iterable, Union, Dict, Tuple

from qtpy import QtWidgets
from qtpy.QtCore import QObject, Slot, Signal, Qt, QRectF
import pyqtgraph as pg
import numpy as np

from pymodaq_data.data import DataRaw, DataFromRoi, Axis, DataToExport, DataCalculated, DataWithAxes
from pymodaq_utils.logger import set_logger, get_module_name
from pymodaq_gui.parameter import utils as putils
from pymodaq_gui.plotting.items.crosshair import Crosshair
from pymodaq_utils import utils

from pymodaq_gui.managers.action_manager import ActionManager
from pymodaq_gui.plotting.data_viewers.viewer import ViewerBase
from pymodaq_gui.plotting.utils.plot_utils import make_dashed_pens, RoiInfo
from pymodaq_gui.managers.roi_manager import ROIManager
from pymodaq_gui.plotting.utils.filter import Filter1DFromCrosshair, Filter1DFromRois
from pymodaq_gui.plotting.widgets import PlotWidget
from pymodaq_gui.plotting.data_viewers.viewer0D import Viewer0D

logger = set_logger(get_module_name(__file__))

PLOT_COLORS = utils.plot_colors


class DataDisplayer(QObject):
    """
    This Object deals with the display of 1D data  on a plotitem
    """

    updated_item = Signal(list)
    labels_changed = Signal(list)

    def __init__(self, plotitem: pg.PlotItem, flip_axes=False, plot_colors=PLOT_COLORS):
        super().__init__()
        self._doxy = False
        self._do_sort = False
        self._do_scatter = False
        self._show_errors = False
        self._flip_axes = flip_axes
        self._plotitem = plotitem
        self._plotitem.addLegend()
        self._plot_items: List[pg.PlotDataItem] = []
        self._boundary_items: List[Tuple[pg.PlotDataItem, pg.PlotDataItem]] = []
        self._fill_items: List[pg.FillBetweenItem] = []
        self._overlay_items: List[pg.PlotDataItem] = []
        self._axis: Axis = None
        self._data: DataWithAxes = None
        self._plot_colors = plot_colors

    @property
    def Ndata(self):
        return len(self._data) if self._data is not None else 0

    def update_colors(self, colors: list):
        self._plot_colors[0:len(colors)] = colors
        self.update_data(self._data, force_update=True)

        self._doxy = False
        self._do_sort = False

    @property
    def legend(self):
        return self._plotitem.legend

    def update_axis(self, axis: Axis):
        self._axis = axis.copy()
        if self._axis.get_data() is None:  # create real data vector once here for subsequent use
            self._axis.create_linear_data(axis.size)

    def get_axis(self) -> Axis:
        return self._axis

    def get_plot_items(self):
        return self._plot_items

    def get_plot_item(self, index: int):
        return self._plot_items[index]

    def update_data(self, data: DataRaw, do_xy=False, sort_data=False, force_update=False,
                    do_scatter=False, show_errors=False):
        if data is not None:
            if data.labels != self.labels or self._show_errors != show_errors:
                force_update = True
            self._data = data
            if len(data) != len(self._plot_items) or force_update:
                self.update_display_items(data, show_errors)
            self.update_plot(do_xy, data, sort_data, do_scatter, show_errors)

    def update_xy(self, do_xy=False):
        self._doxy = do_xy
        self.update_plot(do_xy, sort_data=self._do_sort, scatter=self._do_scatter)

    def update_sort(self, do_sort=False):
        self._do_sort = do_sort
        self.update_plot(self._doxy, sort_data=self._do_sort, scatter=self._do_scatter)

    def update_scatter(self, do_scatter=False):
        self._do_scatter = do_scatter
        self.update_plot(self._doxy, sort_data=self._do_sort, scatter=self._do_scatter)

    def update_errors(self, show_errors=False):
        self._show_errors = show_errors
        self.update_data(self._data, self._doxy, sort_data=self._do_sort, force_update=True,
                         do_scatter=self._do_scatter, show_errors=show_errors)

    def update_plot(self, do_xy=True, data=None, sort_data=False, scatter=False, show_errors=False):
        if data is None:
            data = self._data
        if data is not None:
            if sort_data:
                data = data.sort_data()
                if len(data.axes) == 0:
                    data.create_missing_axes()
            self._doxy = do_xy
            self._do_sort = sort_data
            self._show_errors = show_errors
            self.update_xyplot(do_xy, data)

            if scatter and self.get_plot_items()[0].opts['symbol'] is None:
                if 'symbol_size' in data.extra_attributes:
                    symbol_size = data.symbol_size
                else:
                    symbol_size = 5
                if 'symbol' in data.extra_attributes:
                    symbol = data.symbol
                else:
                    symbol = 'o'
                if 'color' in data.extra_attributes:
                    color = data.color
                else:
                    color = None
                self.plot_with_scatter(True, symbol_size, symbol, color)
            elif not scatter and self.get_plot_items()[0].opts['symbol'] is not None:
                self.plot_with_scatter(False)

    def update_xyplot(self, do_xy=True, dwa: DataWithAxes=None):
        if dwa is None:
            dwa = self._data
        _axis = dwa.get_axis_from_index(0)[0]
        _axis_array = _axis.get_data()
        for ind_data, dat in enumerate(dwa.data):
            if dwa.size > 0:
                if not do_xy:
                    if self._flip_axes:
                        self._plot_items[ind_data].setData(dat, _axis_array)
                    else:
                        self._plot_items[ind_data].setData(_axis_array, dat)
                        if self._show_errors:
                            self._boundary_items[ind_data][0].setData(
                                _axis_array,
                                dat + dwa.get_error(ind_data))
                            self._boundary_items[ind_data][1].setData(
                                _axis_array,
                                dat - dwa.get_error(ind_data))
                else:
                    self._plot_items[ind_data].setData(np.array([]), np.array([]))

        if do_xy:
            for ind, data_array in enumerate(dwa[1:]):
                self._plot_items[ind].setData(dwa.data[0], data_array)
                axis = self._plotitem.getAxis('bottom')
                axis.setLabel(text=dwa.labels[0], units='')
                axis = self._plotitem.getAxis('left')
                axis.setLabel(text=' / '.join(dwa.labels[1:]), units='')
                self.legend.setVisible(False)

        else:
            axis = self._plotitem.getAxis('bottom')
            axis.setLabel(text=_axis.label, units=_axis.units)
            axis = self._plotitem.getAxis('left')
            axis.setLabel(text='', units='')
            self.legend.setVisible(True)

    def plot_with_scatter(self, with_scatter=True, symbol_size=5, symbol='o', color=None):

        for ind, plot_item in enumerate(self.get_plot_items()):
            if color is None:
                color = self._plot_colors[ind]
            if with_scatter:
                pen = None
                symbol_type = symbol
                brush = color

            else:
                pen = color
                symbol_type = None
                brush = None

            plot_item.setPen(pen)
            plot_item.setSymbolBrush(brush)
            plot_item.setSymbol(symbol_type)
            plot_item.setSymbolSize(symbol_size)

    def update_display_items(self, data: DataWithAxes = None, show_errors=False):
        while len(self._plot_items) > 0:
            self._plotitem.removeItem(self._plot_items.pop(0))
            if len(self._boundary_items) > 0:
                b_items = self._boundary_items.pop(0)
                self._plotitem.removeItem(b_items[0])
                self._plotitem.removeItem(b_items[1])
                self._plotitem.removeItem(self._fill_items.pop(0))

        if data is not None:
            for ind in range(len(data)):
                self._plot_items.append(pg.PlotDataItem(pen=self._plot_colors[ind]))
                self._plotitem.addItem(self._plot_items[-1])
                self.legend.addItem(self._plot_items[-1], data.labels[ind])
                if show_errors:
                    self._boundary_items.append((pg.PlotDataItem(pen=list(self._plot_colors[ind]) + [100]),
                                                 pg.PlotDataItem(pen=list(self._plot_colors[ind]) + [100])))
                    self._plotitem.addItem(self._boundary_items[-1][0])
                    self._plotitem.addItem(self._boundary_items[-1][1])
                    self._fill_items.append(pg.FillBetweenItem(*self._boundary_items[-1],
                                                               list(self._plot_colors[ind]) + [100]))
                    self._plotitem.addItem(self._fill_items[-1])

            self.updated_item.emit(self._plot_items)
            self.labels_changed.emit(data.labels)

    @property
    def labels(self):
        if self._data is None:
            return []
        else:
            return self._data.labels

    def legend_items(self):
        return [item[1].text for item in self.legend.items]

    def show_overlay(self, show=True):
        if not show:
            while len(self._overlay_items) > 0:
                self._plotitem.removeItem(self._overlay_items.pop(0))
        else:
            for ind in range(len(self._data)):
                pen = pg.mkPen(color=self._plot_colors[ind], style=Qt.CustomDashLine)
                pen.setDashPattern([10, 10])
                self._overlay_items.append(pg.PlotDataItem(pen=pen))
                self._plotitem.addItem(self._overlay_items[-1])
                if self._do_sort:
                    self._overlay_items[ind].setData(
                        self._data.sort_data().get_axis_from_index(0)[0].get_data(),
                        self._data.sort_data()[ind])
                else:
                    self._overlay_items[ind].setData(
                        self._data.get_axis_from_index(0)[0].get_data(),
                        self._data[ind])



class View1D(ActionManager, QObject):
    def __init__(self, parent_widget: QtWidgets.QWidget = None, show_toolbar=True,
                 no_margins=False, flip_axes=False):
        QObject.__init__(self)
        ActionManager.__init__(self, toolbar=QtWidgets.QToolBar())

        self.no_margins = no_margins
        self.flip_axes = flip_axes

        self.data_displayer: DataDisplayer = None
        self.plot_widget: PlotWidget = None
        self.lineout_widgets: QtWidgets.QWidget = None
        self.lineout_viewers: Viewer0D = None
        self.crosshair: Crosshair = None

        self.roi_target = pg.InfiniteLine(pen='w')
        self.ROIselect = pg.LinearRegionItem(pen='w')

        self.setup_actions()

        self.parent_widget = parent_widget
        if self.parent_widget is None:
            self.parent_widget = QtWidgets.QWidget()
            self.parent_widget.show()

        self.plot_widget = PlotWidget()
        self.roi_manager = ROIManager('1D')
        self.data_displayer = DataDisplayer(self.plotitem, flip_axes=self.flip_axes)
        self.other_data_displayers: Dict[str, DataDisplayer] = {}
        self.setup_widgets()

        self.connect_things()
        self.prepare_ui()

        self.plotitem.addItem(self.roi_target)
        self.plotitem.addItem(self.ROIselect)
        self.roi_target.setVisible(False)
        self.ROIselect.setVisible(False)

        self.show_toolbar = show_toolbar
        if not self.show_toolbar:
            self.splitter_ver.setSizes([0, 1, 0])

    def add_data_displayer(self, displayer_name: str, plot_colors=PLOT_COLORS):
        self.other_data_displayers[displayer_name] = DataDisplayer(self.plotitem, self.flip_axes,
                                                                   plot_colors)

    def remove_data_displayer(self, displayer_name: str):
        displayer = self.other_data_displayers.pop(displayer_name, None)
        if displayer is not None:
            displayer.update_display_items()

    @Slot(int, str, str)
    def add_roi_displayer(self, index, roi_type='', roi_name=''):
        color = self.roi_manager.ROIs[roi_name].color
        self.lineout_viewers.view.add_data_displayer(
            roi_name, make_dashed_pens(color, self.data_displayer.Ndata))

    @Slot(str)
    def remove_roi_displayer(self, roi_name=''):
        self.lineout_viewers.view.remove_data_displayer(roi_name)

    def move_roi_target(self, pos: Iterable[float], **kwargs):
        if not self.roi_target.isVisible():
            self.roi_target.setVisible(True)
        self.roi_target.setPos(pos[0])

    def get_double_clicked(self):
        return self.plot_widget.view.sig_double_clicked

    def display_roi_lineouts(self, roi_dte: DataToExport):
        for roi_name in roi_dte.get_origins('Data0D'):
            self.lineout_viewers.view.display_data(
                roi_dte.get_data_from_name_origin('IntData', roi_name), displayer=roi_name)

    @property
    def axis(self):
        """Get the current axis used to display data"""
        return self.data_displayer.get_axis()

    @property
    def plotitem(self):
        return self.plot_widget.plotItem

    def get_crosshair_signal(self):
        """Convenience function from the Crosshair"""
        return self.crosshair.crosshair_dragged

    def get_crosshair_position(self):
        """Convenience function from the Crosshair"""
        return self.crosshair.get_positions()

    def set_crosshair_position(self, *positions):
        """Convenience function from the Crosshair"""
        self.crosshair.set_crosshair_position(*positions)

    def display_data(self, data: Union[DataWithAxes, DataToExport], displayer: str = None):
        if displayer is None:
            if isinstance(data, DataWithAxes):
                self.set_action_visible('xyplot', len(data) == 2)
                self.data_displayer.update_data(data, self.is_action_checked('xyplot'),
                                                self.is_action_checked('sort'),
                                                do_scatter=self.is_action_checked('scatter'),
                                                show_errors=self.is_action_checked('errors'))
            elif isinstance(data, DataToExport):
                self.set_action_visible('xyplot', len(data[0]) == 2)
                self.data_displayer.update_data(data.pop(0), self.is_action_checked('xyplot'),
                                                self.is_action_checked('sort'),
                                                do_scatter=self.is_action_checked('scatter'),
                                                show_errors=self.is_action_checked('errors'))
                if len(data) > 0:
                    self.data_displayer.add_other_data(data)
        elif displayer in self.other_data_displayers:
            self.other_data_displayers[displayer].update_data(data,
                                                              self.is_action_checked('xyplot'),
                                                              self.is_action_checked('sort'))

    def prepare_ui(self):
        self.show_hide_crosshair(False)

    def do_math(self):
        try:
            if self.is_action_checked('do_math'):
                self.roi_manager.roiwidget.show()
                self.lineout_widgets.show()
            else:
                self.lineout_widgets.hide()
                self.roi_manager.roiwidget.hide()

        except Exception as e:
            logger.exception(str(e))

    @Slot(int, str)
    def update_roi_channels(self, index, roi_type=''):
        """Update the use_channel setting each time a ROI is added"""
        self.roi_manager.update_use_channel(self.data_displayer.labels.copy())

    def setup_widgets(self):
        self.parent_widget.setLayout(QtWidgets.QVBoxLayout())
        if self.no_margins:
            self.parent_widget.layout().setContentsMargins(0, 0, 0, 0)
        splitter_hor = QtWidgets.QSplitter(Qt.Horizontal)
        self.parent_widget.layout().addWidget(splitter_hor)

        self.splitter_ver = QtWidgets.QSplitter(Qt.Vertical)
        splitter_hor.addWidget(self.splitter_ver)
        splitter_hor.addWidget(self.roi_manager.roiwidget)
        self.roi_manager.roiwidget.hide()

        self.splitter_ver.addWidget(self.toolbar)

        self.lineout_widgets = QtWidgets.QWidget()
        self.lineout_viewers = Viewer0D(self.lineout_widgets, show_toolbar=False,
                                        no_margins=True)
        self.lineout_widgets.setContentsMargins(0, 0, 0, 0)
        self.lineout_widgets.hide()

        self.splitter_ver.addWidget(self.plot_widget)
        self.splitter_ver.addWidget(self.lineout_widgets)
        self.roi_manager.viewer_widget = self.plot_widget

        self.crosshair = Crosshair(self.plotitem, orientation='vertical')
        self.show_hide_crosshair()

    def connect_things(self):
        self.connect_action('aspect_ratio', self.lock_aspect_ratio)

        self.connect_action('do_math', self.do_math)
        self.connect_action('scatter', self.data_displayer.update_scatter)
        self.connect_action('xyplot', self.data_displayer.update_xy)
        self.connect_action('sort', self.data_displayer.update_sort)
        self.connect_action('crosshair', self.show_hide_crosshair)
        self.connect_action('overlay', self.data_displayer.show_overlay)
        self.connect_action('errors', self.data_displayer.update_errors)
        self.connect_action('ROIselect', self.show_ROI_select)

        self.roi_manager.new_ROI_signal.connect(self.update_roi_channels)
        self.roi_manager.new_ROI_signal.connect(self.add_roi_displayer)
        self.roi_manager.new_ROI_signal.connect(self.lineout_viewers.get_action('clear').click)
        self.roi_manager.remove_ROI_signal.connect(self.remove_roi_displayer)

        self.roi_manager.color_signal.connect(self.update_colors)
        self.data_displayer.labels_changed.connect(self.roi_manager.update_use_channel)

    def update_colors(self, colors: list):
        for ind, roi_name in enumerate(self.roi_manager.ROIs):
            self.lineout_viewers.update_colors(make_dashed_pens(colors[ind]), displayer=roi_name)

    def show_ROI_select(self):
        self.ROIselect.setVisible(self.is_action_checked('ROIselect'))

    def setup_actions(self):
        self.add_action('do_math', 'Math', 'Calculator', 'Do Math using ROI', checkable=True)
        self.add_action('crosshair', 'Crosshair', 'reset', 'Show data cursor', checkable=True)
        self.add_action('aspect_ratio', 'AspectRatio', 'Zoom_1_1', 'Fix the aspect ratio',
                        checkable=True)
        self.add_action('scatter', 'Scatter', 'Marker', 'Switch between line or scatter plots',
                        checkable=True)
        self.add_action('xyplot', 'XYPlotting', '2d',
                        'Switch between normal or XY representation (valid for 2 channels)',
                        checkable=True,
                        visible=False)
        self.add_action('overlay', 'Overlay', 'overlay', 'Plot overlays of current data',
                        checkable=True)
        self.add_action('errors', 'Errors', 'Statistics2', 'Plot boundaries (~error bars) of '
                                                           'the data',
                        checkable=True)
        self.add_action('sort', 'Sort Data', 'sort_ascend',
                        'Display data in a sorted fashion with respect to axis',
                        checkable=True)
        self.add_action('ROIselect', 'ROI Select', 'Select_24',
                        tip='Show/Hide ROI selection area', checkable=True)
        self.add_action('x_label', 'x:')
        self.add_action('y_label', 'y:')

    def lock_aspect_ratio(self):
        if self.is_action_checked('aspect_ratio'):
            self.plotitem.vb.setAspectLocked(lock=True, ratio=1)
        else:
            self.plotitem.vb.setAspectLocked(lock=False)

    def update_crosshair_data(self, crosshair_dte: DataToExport):
        if len(crosshair_dte) > 0:
            dwa = crosshair_dte[0]
            string = "y="
            for data_array in dwa:
                string += f"{float(data_array[0]):.3e} / "
            self.get_action('y_label').setText(string)
            self.get_action('x_label').setText(f"x={float(dwa.axes[0].get_data()[0]):.3e} ")

    @Slot(bool)
    def show_hide_crosshair(self, show=True):
        self.crosshair.setVisible(show)
        self.set_action_visible('x_label', show)
        self.set_action_visible('y_label', show)
        if self.is_action_checked('crosshair'):
            range = self.plotitem.vb.viewRange()
            self.crosshair.set_crosshair_position(xpos=np.mean(np.array(range[0])))

    def add_plot_item(self, item):
        self.plotitem.addItem(item)


class Viewer1D(ViewerBase):
    """ DataWithAxis of type Data1D can be plotted using this data viewer

    Methods
    -------
    show_data:
        parameter:
        * dwa: a DataWithaxis
        * scatter_dwa: an optional extra DataWithAxis to be plotted with scatter points
          it could define extra_attributes such as symbol: str (to define the symbol layout
          default: 'o') and symbol_size: int (to define the symbol size)

    """

    def __init__(self, parent: QtWidgets.QWidget = None, title='', show_toolbar=True, no_margins=False,
                 flip_axes=False):
        super().__init__(parent=parent, title=title)

        self.view = View1D(self.parent, show_toolbar=show_toolbar, no_margins=no_margins, flip_axes=flip_axes)

        self.filter_from_rois = Filter1DFromRois(self.view.roi_manager)
        self.filter_from_rois.register_activation_signal(self.view.get_action('do_math').triggered)
        self.filter_from_rois.register_target_slot(self.process_roi_lineouts)

        self.filter_from_crosshair = Filter1DFromCrosshair(self.view.crosshair)
        self.filter_from_crosshair.register_activation_signal(self.view.get_action('crosshair').triggered)
        self.filter_from_crosshair.register_target_slot(self.process_crosshair_lineouts)

        self.prepare_connect_ui()

        self._labels = []

    def update_colors(self, colors: List, displayer=None):
        if displayer is None:
            self.view.data_displayer.update_colors(colors)
        elif displayer in self.view.other_data_displayers:
            self.view.other_data_displayers[displayer].update_colors(colors)

    @property
    def roi_manager(self):
        """Convenience method """
        return self.view.roi_manager

    @property
    def roi_target(self) -> pg.InfiniteLine:
        return self.view.roi_target

    def move_roi_target(self, pos: Iterable[float] = None):
        """move a specific read only ROI at the given position on the viewer"""
        self.view.move_roi_target(pos)

    @property
    def crosshair(self):
        """Convenience method """
        return self.view.crosshair

    def set_crosshair_position(self, xpos, ypos=0):
        """Convenience method to set the crosshair positions"""
        self.view.crosshair.set_crosshair_position(xpos=xpos, ypos=ypos)

    def add_plot_item(self, item):
        self.view.add_plot_item(item)

    def process_crosshair_lineouts(self, crosshair_dte: DataToExport):
        self.view.update_crosshair_data(crosshair_dte)
        self.crosshair_dragged.emit(*self.view.crosshair.get_positions())

    def process_roi_lineouts(self, roi_dte: DataToExport):
        self.view.display_roi_lineouts(roi_dte)
        self.measure_data_dict = dict([])
        try:
            self.measure_data_dict = roi_dte.merge_as_dwa('Data0D').to_dict()
            QtWidgets.QApplication.processEvents()
            self.view.roi_manager.settings.child('measurements').setValue(self.measure_data_dict)

            if not self._display_temporary:
                roi_dte_bis = roi_dte.deepcopy()
                for dwa in roi_dte_bis:
                    if dwa.name == 'HorData':
                        dwa.name = f'Hlineout_{dwa.origin}'
                    elif dwa.name == 'IntData':
                        dwa.name = f'Integrated_{dwa.origin}'
                self.data_to_export.append(roi_dte_bis.data)
                self.data_to_export_signal.emit(self.data_to_export)
        except AttributeError:
            pass
        self.ROI_changed.emit()

    def prepare_connect_ui(self):
        self.view.ROIselect.sigRegionChangeFinished.connect(self.selected_region_changed)
        self._data_to_show_signal.connect(self.view.display_data)
        self.roi_manager.roi_changed.connect(self.roi_changed)
        self.roi_manager.roi_value_changed.connect(self.roi_changed)

        self.view.get_crosshair_signal().connect(self.crosshair_changed)
        self.view.get_double_clicked().connect(self.double_clicked)

    def selected_region_changed(self):
        if self.view.is_action_checked('ROIselect'):
            self.roi_select_signal.emit(RoiInfo.info_from_linear_roi(self.view.ROIselect))

    def double_clicked(self, posx, posy=0):
        if self.view.is_action_checked('crosshair'):
            self.view.crosshair.set_crosshair_position(posx)
            self.crosshair_changed()
        self.sig_double_clicked.emit(posx, posy)

    def roi_changed(self):
        self.filter_from_rois.filter_data(self._raw_data)

    def crosshair_changed(self):
        self.filter_from_crosshair.filter_data(self._raw_data)

    def activate_roi(self, activate=True):
        self.view.set_action_checked('do_math', activate)
        self.view.get_action('do_math').triggered.emit(activate)

    @property
    def labels(self):
        return self._labels

    @labels.setter
    def labels(self, labels):
        if labels != self._labels:
            self._labels = labels

    def _show_data(self, data: DataWithAxes, *args, scatter_dwa: DataWithAxes =None,
                   **kwargs):
        self.labels = data.labels
        if len(data.axes) == 0:
            self.get_axis_from_view(data)

        self.view.display_data(data)

        if scatter_dwa is not None:
            if isinstance(scatter_dwa, DataWithAxes):
                if scatter_dwa.name not in self.view.other_data_displayers:
                    self.view.add_data_displayer(scatter_dwa.name, [(255, 0, 0)])
                self.view.other_data_displayers[scatter_dwa.name].update_data(
                    scatter_dwa, do_scatter=True)

        if len(self.view.roi_manager.ROIs) == 0:
            self.data_to_export_signal.emit(self.data_to_export)
        else:
            self.roi_changed()
        if self.view.is_action_checked('crosshair'):
            self.crosshair_changed()

    def get_axis_from_view(self, data: DataWithAxes):
        if self.view.axis is not None:
            data.axes = [self.view.axis]
        else:
            data.create_missing_axes()

    def update_status(self, txt):
        logger.info(txt)


def main():
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    prog = Viewer1D(Form)

    from pymodaq_utils.math_utils import gauss1D

    x = np.linspace(0, 200, 201)
    y1 = gauss1D(x, 75, 25)
    y2 = gauss1D(x, 120, 50, 2)
    tau_half = 27
    tau2 = 100
    x0 = 50
    dx = 20
    ydata_expodec = np.zeros((len(x)))
    ydata_expodec[:50] = 1 * gauss1D(x[:50], x0, dx, 2)
    ydata_expodec[50:] = 1 * np.exp(-(x[50:] - x0) / (tau_half / np.log(2)))  # +1*np.exp(-(x[50:]-x0)/tau2)
    ydata_expodec += 0.1 * np.random.rand(len(x))

    # x = np.sin(np.linspace(0,6*np.pi,201))
    # y = np.sin(np.linspace(0, 6*np.pi, 201)+np.pi/2)
    data = DataRaw('mydata', data=[y1, ydata_expodec, -ydata_expodec, -y1,
                                   y1, ydata_expodec, -ydata_expodec, -y1,
                                   y1, ydata_expodec, -ydata_expodec, -y1,],
                   axes=[Axis('myaxis', 'units', data=x)])

    Form.show()
    prog.view.get_action('do_math').trigger()

    def print_data(data: DataToExport):
        print(data.data)
        print('******')
        print(data.get_data_from_dim('Data1D'))
        print('******')
        print(data.get_data_from_dim('Data0D'))
        print('***********************************')

    #prog.data_to_export_signal.connect(print_data)


    QtWidgets.QApplication.processEvents()
    prog.show_data(data)
    QtWidgets.QApplication.processEvents()
    sys.exit(app.exec_())


def main_unsorted():
    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()
    prog = Viewer1D(widget)

    from pymodaq_utils.math_utils import gauss1D

    x = np.linspace(0, 200, 201)
    xaxis = np.concatenate((x, x[::-1]))
    y = gauss1D(x, 75, 25)
    yaxis = np.concatenate((y, -y))
    data = DataRaw('mydata', data=[yaxis],
                   axes=[Axis('myaxis', 'units', data=xaxis)])
    widget.show()
    prog.show_data(data)

    sys.exit(app.exec_())


def main_random():
    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()
    prog = Viewer1D(widget)

    from pymodaq_utils.math_utils import gauss1D
    x = np.random.randint(201, size=201)
    y1 = gauss1D(x, 75, 25)
    y2 = gauss1D(x, 120, 50, 2)

    QtWidgets.QApplication.processEvents()
    data = DataRaw('mydata', data=[y1, y2],
                   axes=[Axis('myaxis', 'units', data=x, index=0, spread_order=0)],
                   nav_indexes=(0, ))

    widget.show()
    prog.show_data(data)
    QtWidgets.QApplication.processEvents()
    sys.exit(app.exec_())


def main_extra_scatter():
    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()
    prog = Viewer1D(widget)

    from pymodaq_utils.math_utils import gauss1D
    x = np.linspace(0, 200, 201)
    xlow = np.linspace(0, 200, 21)
    y = gauss1D(x, 75, 25)
    ylow = gauss1D(xlow, 75, 25)

    QtWidgets.QApplication.processEvents()
    data = DataRaw('mydata', data=[y],
                   axes=[Axis('myaxis', 'units', data=x, index=0, spread_order=0)],
                   labels=['Initial data'],
                   )
    scatter_dwa = DataRaw('scatter', data=[ylow],
                          axes=[Axis('myaxis', 'units', data=xlow, index=0, spread_order=0)],
                          labels=['subsampled'],
                          symbol='d',
                          symbol_size=18,
                          color='b')

    widget.show()
    prog.show_data(data, scatter_dwa=scatter_dwa)
    QtWidgets.QApplication.processEvents()
    sys.exit(app.exec_())


def main_errors():
    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()
    prog = Viewer1D(widget)

    from pymodaq_utils.math_utils import gauss1D
    x = np.linspace(0, 200, 201)
    y1 = gauss1D(x, 75, 25)
    y2 = gauss1D(x, 120, 50, 2)

    QtWidgets.QApplication.processEvents()
    data = DataRaw('mydata', data=[y1, y2],
                   axes=[Axis('myaxis', 'units', data=x, index=0, spread_order=0)],
                   errors=(np.random.random_sample(x.shape),
                           np.random.random_sample(x.shape)))


    widget.show()
    prog.show_data(data)
    QtWidgets.QApplication.processEvents()
    sys.exit(app.exec_())


def main_view1D():
    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()
    prog = View1D(widget)
    widget.show()
    sys.exit(app.exec_())


def main_nans():
    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()
    prog = Viewer1D(widget)

    from pymodaq_utils.math_utils import gauss1D

    x = np.linspace(0, 200, 201)
    y = gauss1D(x, 75, 25)

    y[100:150] = np.nan
    data = DataRaw('mydata', data=[y],
                   axes=[Axis('myaxis', 'units', data=x)])

    widget.show()
    prog.show_data(data)

    sys.exit(app.exec_())


if __name__ == '__main__':  # pragma: no cover
    main()
    # main_random()
    #main_errors()
    #main_extra_scatter()
    #main_view1D()
    #main_nans()



# -*- coding: utf-8 -*-
"""
Created the 15/11/2022

@author: Sebastien Weber
"""
from typing import Tuple
import os
from collections import OrderedDict
from typing import List
import warnings
import logging
import webbrowser
import numpy as np
from pathlib import Path
from packaging import version as version_mod

from qtpy import QtGui, QtCore
from qtpy.QtCore import Qt, QObject, Signal, QByteArray
from qtpy import QtWidgets

from pymodaq_utils.logger import set_logger, get_module_name
from pymodaq_utils.config import Config
from pymodaq_utils import utils

from pymodaq_data.h5modules import data_saving
from pymodaq_data.h5modules.exporter import ExporterFactory
from pymodaq_data.h5modules.browsing import H5BrowserUtil

from pymodaq_gui.parameter import ioxml
from pymodaq_gui.utils.widgets.tree_layout import TreeLayout
from pymodaq_gui.utils.utils import h5tree_to_QTree, pngbinary2Qlabel
from pymodaq_gui.utils.file_io import select_file, select_file_filter
from pymodaq_gui.plotting.data_viewers.viewerND import ViewerND
from pymodaq_gui.managers.action_manager import ActionManager
from pymodaq_gui.managers.parameter_manager import ParameterManager
from pymodaq_gui.messenger import messagebox


config = Config()
logger = set_logger(get_module_name(__file__))


class View(QObject):
    item_clicked_sig = Signal(object)
    item_double_clicked_sig = Signal(object)
    
    def __init__(self, widget: QtWidgets.QWidget, settings_tree, settings_attributes_tree):
        super().__init__()
        self.parent_widget = widget
        self.h5file_tree: TreeLayout = None

        self._viewer_widget: QtWidgets.QWidget = None
        self._text_list: QtWidgets.QListWidget = None
        self._pixmap_widget: QtWidgets.QWidget = None

        self.setup_ui(settings_tree, settings_attributes_tree)

    def setup_ui(self, settings_tree, settings_attributes_tree):
        layout = QtWidgets.QGridLayout()

        v_splitter = QtWidgets.QSplitter(Qt.Vertical)
        v_splitter2 = QtWidgets.QSplitter(Qt.Vertical)
        h_splitter = QtWidgets.QSplitter(Qt.Horizontal)

        widget = QtWidgets.QWidget()
        # self.ui.h5file_tree = TreeLayout(Form,col_counts=2,labels=["Node",'Pixmap'])
        self.h5file_tree = TreeLayout(widget, col_counts=1, labels=["Node"])
        self.h5file_tree.tree.setMinimumWidth(300)

        self.h5file_tree.item_clicked_sig.connect(self.item_clicked_sig.emit)
        self.h5file_tree.item_double_clicked_sig.connect(self.item_double_clicked_sig.emit)
        
        v_splitter.addWidget(widget)
        v_splitter.addWidget(settings_attributes_tree)

        h_splitter.addWidget(v_splitter)
        self._pixmap_widget = QtWidgets.QWidget()
        self._pixmap_widget.setMaximumHeight(100)
        v_splitter2.addWidget(self._pixmap_widget)

        v_splitter2.addWidget(settings_tree)
        self._text_list = QtWidgets.QListWidget()

        v_splitter2.addWidget(self._text_list)
        h_splitter.addWidget(v_splitter2)
        self._viewer_widget = QtWidgets.QWidget()
        h_splitter.addWidget(self._viewer_widget)
        layout.addWidget(h_splitter)
        self.parent_widget.setLayout(layout)

    def current_node_path(self):
        return self.h5file_tree.current_node_path()

    def add_actions(self, actions: List[QtWidgets.QAction]):
        for action in actions:
            self.h5file_tree.tree.addAction(action)
          
    @property  
    def viewer_widget(self):
        return self._viewer_widget

    @property
    def text_list(self):
        return self._text_list

    @property
    def pixmap_widget(self):
        return self._pixmap_widget

    def clear(self):
        self.h5file_tree.tree.clear()

    def add_base_item(self, base_tree_item):
        self.h5file_tree.tree.addTopLevelItem(base_tree_item)

    def add_widget_to_tree(self, pixmap_items):
        for item in pixmap_items:
            widget = QtWidgets.QWidget()

            vLayout = QtWidgets.QVBoxLayout()
            label1D = QtWidgets.QLabel()
            bytes = QByteArray(item['node'].attrs['pixmap1D'])
            im1 = QtGui.QImage.fromData(bytes)
            a = QtGui.QPixmap.fromImage(im1)
            label1D.setPixmap(a)

            label2D = QtWidgets.QLabel()
            bytes = QByteArray(item['node'].attrs['pixmap2D'])
            im2 = QtGui.QImage.fromData(bytes)
            b = QtGui.QPixmap.fromImage(im2)
            label2D.setPixmap(b)

            vLayout.addWidget(label1D)
            vLayout.addwidget(label2D)
            widget.setLayout(vLayout)
            self.h5file_tree.tree.setItemWidget(item['item'], 1, widget)


class H5Browser(QObject, ActionManager):
    """UI used to explore h5 files, plot and export subdatas

    Parameters
    ----------
    parent: QtWidgets container
        either a QWidget or a QMainWindow
    h5file: h5file instance
        exact type depends on the backend
    h5file_path: str or Path
        if specified load the corresponding file, otherwise open a select file dialog
    backend: str
        either 'tables, 'h5py' or 'h5pyd'

    See Also
    --------
    H5Backend, H5Backend
    """
    data_node_signal = Signal(str)  # the path of a node where data should be monitored, displayed...
    # whatever use from the caller
    status_signal = Signal(str)

    def __init__(self, parent: QtWidgets.QMainWindow, h5file=None, h5file_path=None, backend='tables'):
        QObject.__init__(self)
        # toolbar = QtWidgets.QToolBar()
        ActionManager.__init__(self)  # , toolbar=toolbar)

        if not isinstance(parent, QtWidgets.QMainWindow):
            raise Exception('no valid parent container, expected a QMainWindow')

        self.main_window = parent
        self.parent_widget = QtWidgets.QWidget()
        self.main_window.setCentralWidget(self.parent_widget)
        #self.main_window.addToolBar(self.toolbar)

        self.current_node_path = None

        self.settings_attributes = ParameterManager()
        self.settings = ParameterManager()

        # construct the UI interface
        self.view = View(self.parent_widget, settings_tree=self.settings.settings_tree,
                         settings_attributes_tree=self.settings_attributes.settings_tree)
        self.view.item_clicked_sig.connect(self.show_h5_attributes)
        self.view.item_double_clicked_sig.connect(self.show_h5_data)
        self.hyper_viewer = ViewerND(self.view.viewer_widget)

        self.setup_actions()
        self.setup_menu()
        self.connect_things()

        # construct the h5 interface and load the file (or open a select file message)
        self.h5utils = H5BrowserUtil(backend=backend)
        self.data_loader = None
        self.load_file(h5file, h5file_path)

    def connect_things(self):
        self.connect_action('export', self.export_data)
        self.connect_action('comment', self.add_comments)
        self.connect_action('load', lambda: self.load_file(None, None))
        self.connect_action('save', self.save_file)
        self.connect_action('quit', self.quit_fun)
        self.connect_action('about', self.show_about)
        self.connect_action('help', self.show_help)
        self.connect_action('log', self.show_log)

        self.connect_action('plot_node', lambda: self.get_node_and_plot(False))
        self.connect_action('plot_nodes', lambda: self.get_node_and_plot(False, True))
        self.connect_action('plot_node_with_bkg', lambda: self.get_node_and_plot(True))
        self.connect_action('plot_nodes_with_bkg', lambda: self.get_node_and_plot(True, True))

        self.status_signal.connect(self.add_log)

    def get_node_and_plot(self, with_bkg, plot_all=False):
        self.show_h5_data(item=None, with_bkg=with_bkg, plot_all=plot_all)

    def load_file(self, h5file=None, h5file_path=None):
        if h5file is None:
            if h5file_path is None:
                h5file_path = select_file(save=False, ext=['h5', 'hdf5'])
            if Path(h5file_path).is_file():
                if self.h5utils.isopen():
                    self.h5utils.close_file()

                self.h5utils.open_file(h5file_path, 'r+')
            else:
                return
        else:
            self.h5utils.h5file = h5file

        self.data_loader = data_saving.DataLoader(self.h5utils)
        self.check_version()
        self.populate_tree()
        self.view.h5file_tree.expand_all()

    def setup_menu(self):
        menubar = self.main_window.menuBar()
        file_menu = menubar.addMenu('File')
        self.affect_to('load', file_menu)
        self.affect_to('save', file_menu)
        file_menu.addSeparator()
        self.affect_to('quit', file_menu)

        help_menu = menubar.addMenu('?')
        self.affect_to('about', help_menu)
        self.affect_to('help', help_menu)
        self.affect_to('log', help_menu)

    def setup_actions(self):
        self.add_action('export', 'Export as', 'SaveAs', tip='Export node content (and children) as ',
                        toolbar=self.toolbar)
        self.add_action('comment', 'Add Comment', 'properties', tip='Add comments to the node',
                        toolbar=self.toolbar)
        self.add_action('plot_node', 'Plot Node', 'color', tip='Plot the current node',
                        toolbar=self.toolbar)
        self.add_action('plot_nodes', 'Plot Nodes', 'color', tip='Plot all nodes hanging from the same parent',
                        toolbar=self.toolbar)
        self.add_action('plot_node_with_bkg', 'Plot Node With Bkg', 'color', tip='Plot the current node with background'
                                                                                 ' substraction if possible',
                        toolbar=self.toolbar)
        self.add_action('plot_nodes_with_bkg', 'Plot Nodes With Bkg', 'color', tip='Plot all nodes hanging from'
                                                                                   ' the same parent with background'
                                                                                   ' substraction if possible',
                        toolbar=self.toolbar)
        self.view.add_actions([self.get_action('export'), self.get_action('comment'),
                               self.get_action('plot_node'), self.get_action('plot_nodes'),
                               self.get_action('plot_node_with_bkg'),
                               self.get_action('plot_nodes_with_bkg')])

        self.add_action('load', 'Load File', 'Open', tip='Open a new file')
        self.add_action('save', 'Save File as', 'SaveAs', tip='Save as another file')
        self.add_action('quit', 'Quit the application', 'Exit', tip='Quit the application')
        self.add_action('about', 'About', tip='About')
        self.add_action('help', 'Help', 'Help', tip='Show documentation', shortcut=QtCore.Qt.Key_F1)
        self.add_action('log', 'Show Log', 'information2', tip='Open Log')

    def check_version(self):
        """Check version of PyMoDAQ to assert if file is compatible or not with the current version of the Browser"""
        if 'pymodaq_version' in self.h5utils.root().attrs.attrs_name:
            if version_mod.parse(self.h5utils.root().attrs['pymodaq_version']) < version_mod.parse('4.0.0a0'):
                msg_box = messagebox(severity='warning', title='Invalid version',
                                     text=f"Your file has been saved using PyMoDAQ "
                                          f"version {self.h5utils.root().attrs['pymodaq_version']} "
                                          f"while you're using version: {utils.get_version()}\n"
                                          f"Please create and use an adapted environment to use this"
                                          f" version (up to 3.x.y):\n"
                                          f"pip install pymodaq==3.x.y")
                self.quit_fun()

    def add_comments(self, status: bool, comment=''):
        """Add comments to a node

        Parameters
        ----------
        status: bool
        comment: str
            The comment to be added in a comment attribute to the current node path

        See Also
        --------
        current_node_path
        """
        try:
            self.current_node_path = self.get_tree_node_path()
            node = self.h5utils.get_node(self.current_node_path)
            if 'comments' in node.attrs.attrs_name:
                tmp = node.attrs['comments']
            else:
                tmp = ''
            if comment == '':
                text, res = QtWidgets.QInputDialog.getMultiLineText(None, 'Enter comments', 'Enter comments here:', tmp)
                if res and text != '':
                    comment = text
                node.attrs['comments'] = comment
            else:
                node.attrs['comments'] = tmp + comment

            self.h5utils.flush()

        except Exception as e:
            logger.exception(str(e))

    def get_tree_node_path(self):
        """Get the node path of the currently selected node in the UI"""
        return self.view.current_node_path()

    def export_data(self):
        """Opens a dialog to export data

        See Also
        --------
        H5BrowserUtil.export_data
        """
        try:
            file_filter = ExporterFactory.get_file_filters()
            file, selected_filter = select_file_filter(save=True, filter=file_filter)
            self.current_node_path = self.get_tree_node_path()
            if file != '':
                self.h5utils.export_data(self.current_node_path, str(file), selected_filter)

        except Exception as e:
            logger.exception(str(e))

    def save_file(self, filename=None):

        if filename is None:
            filename = select_file(save=True, ext='txt')
        if filename != '':
            self.h5utils.save_file(filename)

    def quit_fun(self):
        """
        """
        try:
            self.h5utils.close_file()
            if self.main_window is None:
                self.parent_widget.close()
            else:
                self.main_window.close()
        except Exception as e:
            logger.exception(str(e))

    def show_about(self):
        splash_path = os.path.join(os.path.split(os.path.split(__file__)[0])[0], 'splash.png')
        splash = QtGui.QPixmap(splash_path)
        self.splash_sc = QtWidgets.QSplashScreen(splash, QtCore.Qt.WindowStaysOnTopHint)
        self.splash_sc.setVisible(True)
        self.splash_sc.showMessage(f"PyMoDAQ version {utils.get_version()}\n"
                                   f"Modular Acquisition with Python\nWritten by Sébastien Weber",
                                   QtCore.Qt.AlignRight, QtCore.Qt.white)

    @staticmethod
    def show_log():
        webbrowser.open(logging.getLogger('pymodaq').handlers[0].baseFilename)

    @staticmethod
    def show_help():
        QtGui.QDesktopServices.openUrl(QtCore.QUrl("http://pymodaq.cnrs.fr"))

    @staticmethod
    def add_log(txt):
        logger.info(txt)

    def show_h5_attributes(self, item=None):
        try:
            self.current_node_path = self.get_tree_node_path()

            attr_dict, settings, scan_settings, pixmaps = self.h5utils.get_h5_attributes(self.current_node_path)

            for child in self.settings_attributes.settings.children():
                child.remove()
            params = []
            for attr in attr_dict:
                params.append({'title': attr, 'name': attr, 'type': 'str', 'value': attr_dict[attr],
                               'readonly': True})
            self.settings_attributes.settings.addChildren(params)

            if settings is not None:
                for child in self.settings.settings.children():
                    child.remove()
                QtWidgets.QApplication.processEvents()  # so that the tree associated with settings updates
                params = ioxml.XML_string_to_parameter(settings)
                self.settings.settings.addChildren(params)

            if scan_settings is not None:
                params = ioxml.XML_string_to_parameter(scan_settings)
                self.settings.settings.addChildren(params)

            if pixmaps == []:
                self.view.pixmap_widget.setVisible(False)
            else:
                self.view.pixmap_widget.setVisible(True)
                self.show_pixmaps(pixmaps)

        except Exception as e:
            logger.exception(str(e))

    def show_pixmaps(self, pixmaps=[]):
        if self.view.pixmap_widget.layout() is None:
            self.view.pixmap_widget.setLayout(QtWidgets.QHBoxLayout())
        while 1:
            child = self.view.pixmap_widget.layout().takeAt(0)
            if not child:
                break
            child.widget().deleteLater()
            QtWidgets.QApplication.processEvents()
        labs = []
        for pix in pixmaps:
            labs.append(pngbinary2Qlabel(pix))
            self.view.pixmap_widget.layout().addWidget(labs[-1])

    def show_h5_data(self, item, with_bkg=False, plot_all=False):
        """

        Parameters
        ----------
        item
        with_bkg
        plot_all

        Returns
        -------

        """
        try:
            if item is None:
                self.current_node_path = self.get_tree_node_path()
            self.show_h5_attributes()
            node = self.h5utils.get_node(self.current_node_path)
            self.data_node_signal.emit(self.current_node_path)

            if 'data_type' in node.attrs and node.attrs['data_type'] == 'strings':
                self.view.text_list.clear()
                for txt in node.read():
                    self.view.text_list.addItem(txt)
            elif 'data_type' in node.attrs:
                data_with_axes = self.data_loader.load_data(node, with_bkg=with_bkg, load_all=plot_all)
                self.hyper_viewer.show_data(data_with_axes, force_update=True)

        except Exception as e:
            logger.exception(str(e))

    def populate_tree(self):
        """
            | Init the ui-tree and store data into calling the h5_tree_to_Qtree convertor method

            See Also
            --------
            h5tree_to_QTree, update_status
        """
        try:
            if self.h5utils.h5file is not None:
                self.view.clear()
                base_node = self.h5utils.root()
                base_tree_item, pixmap_items = h5tree_to_QTree(base_node)
                self.view.add_base_item(base_tree_item)
                self.view.add_widget_to_tree(pixmap_items)

        except Exception as e:
            logger.exception(str(e))


def browse_data(fname=None, ret_all=False, message=None) -> Tuple[data_saving.DataWithAxes, str, str]:
    """Browse data present in any h5 file using the H5Browser within a dialog window
    when the user has selected a given node, return its content

    Parameters
    ----------
    fname: str
    ret_all: bool
    message: str

    Returns
    -------
    data: DataWithAxes
    if argument ret_all is True, returns also:
    fname: the file name
    node_path: hte path of the selected node within the H5 file tree

    """
    if fname is None:
        fname = str(select_file(start_path=config('data_saving', 'h5file', 'save_path'), save=False, ext='h5'))

    if type(fname) != str:
        try:
            fname = str(fname)
        except Exception:
            raise Exception('filename in browse data is not valid')
    if fname != '':
        (root, ext) = os.path.splitext(fname)
        if not ('h5' in ext or 'hdf5' in ext):
            warnings.warn('This is not a PyMODAQ h5 file, there could be issues', Warning)

        form = QtWidgets.QMainWindow()
        browser = H5Browser(form, h5file_path=fname)
        dataloader = data_saving.DataLoader(browser.h5utils)
        dialog = QtWidgets.QDialog()
        vlayout = QtWidgets.QVBoxLayout()

        vlayout.addWidget(form)
        dialog.setLayout(vlayout)
        buttonBox = QtWidgets.QDialogButtonBox(parent=dialog)

        buttonBox.addButton('OK', buttonBox.AcceptRole)
        buttonBox.accepted.connect(dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(dialog.reject)
        vlayout.addWidget(buttonBox)

        dialog.setWindowTitle('Select a data node in the tree')
        if message is None or not isinstance(message, str):
            dialog.setWindowTitle('Select a data node in the tree')
        else:
            dialog.setWindowTitle(message)
        res = dialog.exec()

        if res == dialog.Accepted:
            node_path = browser.current_node_path
            data = dataloader.load_data(node_path, with_bkg=True)
        else:
            data = None
            node_path = None

        browser.h5utils.close_file()

        if ret_all:
            return data, fname, node_path
        else:
            return data
    return None, '', ''
